## import standard libraries
import abc
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional
# import locals
from ogd.core.generators.Generator import Generator, GeneratorParameters
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.utils.typing import Map

## @class Model
#  Abstract base class for session-level Wave Detectors.
#  Models only have one public function, called Eval.
#  The Eval function takes a list of row data, computes some statistic, and returns a list of results.
#  If the model works on Detectors from session data, it should calculate one result for each row (each row being a session).
#  If the model works on a raw list of recent events, it should calculate a single result (each row being an event).
class Detector(Generator):
#TODO: use a dirty bit so we only run the GetValue function if we've received an event or Detector since last calculation

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _trigger_condition(self) -> bool:
        pass

    @abc.abstractmethod
    def _trigger_event(self) -> DetectorEvent:
        pass

    # *** BUILT-INS & PROPERTIES *8jr47t*

    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None]):
        super().__init__(params=params)
        self._callback        = trigger_callback
        self._saw_first_event : bool            = False
        # Set up variables for default values of DetectorEvent elements
        self._triggering_event: Event

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """Base function for getting any features a second-order feature depends upon.
        By default, no dependencies.
        Any feature intented to be second-order should override this function.

        :return: _description_
        :rtype: List[str]
        """
        return []

    # *** PUBLIC STATICS ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Detector.

        Overridden from Extractor's version of the function, only makes Detector mode supported.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.DETECTOR]

    # *** PUBLIC METHODS ***

    def UpdateFromEvent(self, event:Event):
        if self._validateEvent(event=event):
            self._updateFromEvent(event=event)
            if self._trigger_condition():
                self._triggering_event = event
                _new_event = self._trigger_event()
                self._callback(_new_event)

    def GenerateEvent(self, event_name:str,              event_data:Map,
                      session_id:Optional[str]=None,     app_id:Optional[str]=None,
                      timestamp:Optional[datetime]=None, time_offset:Optional[timedelta]=None,
                      app_version:Optional[str]=None,    log_version:Optional[str]=None,
                      user_id:Optional[str] = None,      user_data:Optional[Map] = None,
                      game_state:Optional[Map]=None,     event_sequence_index:Optional[int]=None):
        return DetectorEvent(
            session_id = session_id   or self._triggering_event.SessionID,
            app_id     = app_id       or self._triggering_event.AppID,
            event_name = event_name,  # no default, must be provided
            event_data = event_data,  # no default, must be provided
            timestamp  = timestamp    or self._triggering_event.Timestamp,
            time_offset= time_offset  or self._triggering_event.TimeOffset,
            app_version= app_version  or self._triggering_event.AppVersion,
            log_version= log_version  or self._triggering_event.LogVersion,
            user_id    = user_id      or self._triggering_event.UserID,
            user_data  = user_data    or self._triggering_event.UserData,
            game_state = game_state   or self._triggering_event.GameState,
            event_sequence_index = event_sequence_index or self._triggering_event.EventSequenceIndex
        )

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import json
from typing import Any, Final, List, Optional
import json
from datetime  import timedelta, datetime
# import local files
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature





class PlayTime(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    IDLE_TIME_THRESHOLD : Final[timedelta] = timedelta(seconds=60)


    def __init__(self, params:GeneratorParameters, threshold: int):
        super().__init__(params=params)
        self._last_event_timestamp : Optional[datetime] = None
        self._first_event_timestamp : Optional[datetime] = None
        
        
        self._idle_time : timedelta = timedelta(0)
        self._count : int = 0
        self._last_timestamp : Optional[datetime] = None
        self._threshold : timedelta = timedelta(seconds=threshold)
        self._total_time : Optional[timedelta] = None

        ##bugfixing
        self._user_log = None

    @staticmethod
    def defaultThreshold():
        return PlayTime.IDLE_TIME_THRESHOLD

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        

        if(not self._first_event_timestamp):
            self._first_event_timestamp = event.Timestamp

        
        if not self._last_event_timestamp:
            self._last_event_timestamp = event.Timestamp
            return
        if self._last_event_timestamp is not None:
            time_since_last = event.Timestamp - self._last_event_timestamp
            if time_since_last > PlayTime.IDLE_TIME_THRESHOLD:
                self._idle_time += time_since_last
                self._count += 1
        else:
            raise ValueError("In IdleState, last timestamp is None!")
        self._last_event_timestamp = event.Timestamp

        if(not event.UserID is None):
            self._user_log = event.UserID

        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """

        #exception handling for empty sessions

        try:
            total_time : timedelta = self._last_event_timestamp - self._first_event_timestamp
            play_time : timedelta = total_time - self._idle_time
        except:
            total_time = timedelta(0)
            play_time  = timedelta(0)


        return [total_time, play_time, self._idle_time]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Active", "Idle"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



# import libraries
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature

class RegionName(PerCountFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        regions = ['arctic', 'coral', 'bayou', 'kelp', 'other']
        self.count = 0
        self.region = regions[self.CountIndex]
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _validateEventCountIndex(self, event:Event):
        pass

    def _updateFromEvent(self, event:Event) -> None:
        pass

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self.region]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import standard libraries
from enum import IntEnum

class ExportMode(IntEnum):
    SESSION = 1
    PLAYER = 2
    POPULATION = 3
    EVENTS = 4
    DETECTORS = 5

    def __str__(self):
        return self.name



# import libraries
from typing import List, Optional, Type
# import locals
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

class LegacyLoader(GeneratorLoader):
    def __init__(self, player_id:str, session_id:str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)

    def GetFeatureClass(self, feature_type:str) -> Optional[Type[Feature]]:
        return None

    def _validateMode(self, feature_type) -> bool:
        """Overridden version of _validateMode.
        For LegacyLoader, we never care to check modes, since feature is a monolith and should just go for all versions.

        :param feature_type: _description_
        :type feature_type: _type_
        :return: _description_
        :rtype: bool
        """
        return True



import logging
from datetime import datetime
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class SnippetsCollected(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._snippet_ids   : List[str] = []
        self._snippet_quals : List[str] = []
        self._snippet_types : List[str] = []


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["snippet_received"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "snippet_received":
            self._snippet_ids.append(event.EventData.get("snippet_id", "SNIPPET ID NOT FOUND"))
            self._snippet_quals.append(event.EventData.get("snippet_quality", "SNIPPET QUALITY NOT FOUND"))
            self._snippet_types.append(event.EventData.get("snippet_type", "SNIPPET TYPE NOT FOUND"))

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._snippet_ids, self._snippet_quals, self._snippet_types]

    # *** Optionally override public functions. ***

    def Subfeatures(self) -> List[str]:
        return ["Qualities", "Types"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER, ExtractionMode.SESSION]



# import standard libraries
import abc
import logging
from typing import Any, Dict, Optional, Type
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.schemas.configs.data_sources.DataSourceSchema import DataSourceSchema
from ogd.core.utils.Logger import Logger

class BigQuerySchema(DataSourceSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any], fallbacks:Dict[str, Any]={}):
        self._project_id : str
        self._credential : Optional[str]

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Game Source config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "PROJECT_ID" in all_elements.keys():
            self._project_id = BigQuerySchema._parseProjectID(all_elements["PROJECT_ID"])
        elif "PROJECT_ID" in fallbacks.keys():
            self._project_id = BigQuerySchema._parseProjectID(fallbacks["PROJECT_ID"])
        else:
            self._project_id = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'DATASET_ID' element; defaulting to dataset_id={self._project_id}", logging.WARN)
        if "PROJECT_KEY" in all_elements.keys():
            self._credential = BigQuerySchema._parseCredential(all_elements["PROJECT_KEY"])
        elif "PROJECT_KEY" in fallbacks.keys():
            self._credential = BigQuerySchema._parseCredential(fallbacks["PROJECT_KEY"])
        else:
            self._credential = None
            Logger.Log(f"{name} config does not have a 'PROJECT_KEY' element; defaulting to credential=None", logging.WARN)

        _used = {"PROJECT_ID", "DATASET_ID", "PROJECT_KEY"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def ProjectID(self) -> str:
        return self._project_id

    @property
    def Credential(self) -> Optional[str]:
        return self._credential

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name}: `{self.AsConnectionInfo}` ({self.Type})"
        return ret_val

    @property
    def AsConnectionInfo(self) -> str:
        ret_val : str = f"{self.ProjectID}"
        return ret_val

    @staticmethod
    def _parseProjectID(project_id) -> str:
        ret_val : str
        if isinstance(project_id, str):
            ret_val = project_id
        else:
            ret_val = str(project_id)
            Logger.Log(f"Data Source project ID was unexpected type {type(project_id)}, defaulting to str(project_id)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseCredential(credential) -> str:
        ret_val : str
        if isinstance(credential, str):
            ret_val = credential
        else:
            ret_val = str(credential)
            Logger.Log(f"Game Source credential type was unexpected type {type(credential)}, defaulting to str(credential)={ret_val}.", logging.WARN)
        return ret_val




## import standard libraries
import logging
import math
import re
import typing
## import local libraries
from models.SequenceModel import SequenceModel

## @class LogisticModel
class NthEventModel(SequenceModel):
    def __init__(self, n: int, levels: typing.List[int] = []):
        super().__init__(levels)
        self._n = n

    def _eval(self, rows: typing.List[typing.Dict[str, typing.Any]]) -> typing.Any:
        return rows[self._n]


# import standard libraries
import logging
from typing import Any, Dict, Union
# import local files
from ogd.core.schemas.games.FeatureSchema import FeatureSchema
from ogd.core.utils.Logger import Logger

class PerCountSchema(FeatureSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._count  : Union[int, str]
        self._prefix : str

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Per-count Feature config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "count" in all_elements.keys():
            self._count = PerCountSchema._parseCount(all_elements["count"])
        else:
            self._count = 0
            Logger.Log(f"{name} config does not have a 'count' element; defaulting to count={self._count}", logging.WARN)
        if "prefix" in all_elements.keys():
            self._prefix = PerCountSchema._parsePrefix(all_elements['prefix'])
        else:
            self._prefix = "pre"
            Logger.Log(f"{name} config does not have a 'prefix' element; defaulting to prefix='{self._prefix}'", logging.WARN)

        _leftovers = { key : val for key,val in all_elements.items() if key not in {"count", "prefix"} }
        super().__init__(name=name, all_elements=_leftovers)

    @property
    def Count(self) -> Union[int, str]:
        return self._count

    @property
    def Prefix(self) -> str:
        return self._prefix

    @property
    def AsMarkdown(self) -> str:
        ret_val   : str

        ret_val = f"**{self.Name}** : *{self.ReturnType}*, *Per-count feature* {' (disabled)' if not len(self.Enabled) > 0 else ''}  \n{self.Description}  \n"
        if len(self.Subfeatures) > 0:
            ret_val += "*Sub-features*:  \n\n" + "\n".join([subfeature.AsMarkdown for subfeature in self.Subfeatures.values()])
        if len(self.NonStandardElements) > 0:
            ret_val += "*Other elements*:  \n\n" + "\n".join([f"{elem_name} : {elem}" for elem_name,elem in self.NonStandardElements.items()])
        return ret_val

    @staticmethod
    def _parseCount(count) -> Union[int, str]:
        ret_val : Union[int, str]
        if isinstance(count, int):
            ret_val = count
        elif isinstance(count, str):
            ret_val = count
        else:
            ret_val = 0
            Logger.Log(f"Extractor count was unexpected type {type(count)}, defaulting to count=0.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePrefix(prefix) -> str:
        ret_val : str
        if isinstance(prefix, str):
            ret_val = prefix
        else:
            ret_val = str(prefix)
            Logger.Log(f"Extractor prefix was unexpected type {type(prefix)}, defaulting to str(prefix).", logging.WARN)
        return ret_val



from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class SucceededThroughFailure(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.failure_count: int = 0
        self.win_after_failure: bool = False

    # Implement abstract functions
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["lose_game", "win_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        event_type = event.EventName
        if event_type == "lose_game":
            self.failure_count += 1
        elif event_type == "win_game" and self.failure_count > 0:
            self.win_after_failure = True

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self.failure_count > 0:
            return [self.win_after_failure]
        else:
            return [None]

    # Optionally override public functions
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



from typing import Any, List
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class PhasesReached(SessionFeature):

    def __init__(self, params: GeneratorParameters):
        self.phases_reached = set()
        super().__init__(params=params)

    @classmethod
    def _getEventDependencies(cls, mode: ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _getFeatureDependencies(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _extractFromEvent(self, event: Event) -> None:
        region = event.GameState.get("region")
        if region:
            self.phases_reached.add(region)

    def _extractFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [list(self.phases_reached)]



# import libraries
from typing import Any, List
from ogd.core.models import Event
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


"""class StoryAlignmentSequence(PerLevelFeature):
    def __init__(self, params: GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._story_alignment_sequence = []

    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["story_updated"]

    @classmethod    
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        self._story_alignment_sequence.append(event.EventData["story_alignment"])

    def _updateFromFeatureData(self, feature: FeatureData):
        return []

    def _getFeatureValues(self) -> List[Any]:
        return [self._story_alignment_sequence]

    def Subfeatures(self) -> List[str]:
        return []

    @classmethod
    def AvailableModes(cls) -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER, ExtractionMode.SESSION]"""


class StoryScoreSequence(PerLevelFeature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self._story_score_sequence = []

    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["story_updated"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        self._story_score_sequence.append(event.event_data["story_score"])

    def _updateFromFeatureData(self, feature: FeatureData) -> None:
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._story_score_sequence]

    def Subfeatures(self) -> List[str]:
        return []

    @classmethod
    def AvailableModes(cls) -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER, ExtractionMode.SESSION]




# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from datetime import datetime, timedelta
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature

class SnowBallDuration(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None
        self._argument_start_time : Optional[datetime] = None
        self._prev_timestamp = None
        self._time = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["push_snowball"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID
            # if we jumped to a new session, we only want to count time up to last event, not the time between sessions.
            if self._argument_start_time and self._prev_timestamp:
                self._time += (self._prev_timestamp - self._argument_start_time).total_seconds()
                self._argument_start_time = event.Timestamp

        elif event.EventName == "push_snowball" and self._argument_start_time is not None:
            self._time = (event.Timestamp - self._argument_start_time).total_seconds()
            self._argument_start_time = None
            return
        self._prev_timestamp = event.Timestamp
    
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    


from typing import List, Optional, Dict, Any
import pandas as pd
from models.SequenceModel import SequenceModel

FB_DEATH_ENUM = 18

## @class DeathCountModel
# Returns the number of deaths in the game
# @param levels: Levels applicable for model
class DeathCountModel(SequenceModel):

    def __init__(self, levels=[]):
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> List[int]:
        assert events
        dh_cnt = 0
        for row in events:
            if row['event_custom'] == FB_DEATH_ENUM:
                dh_cnt += 1
        return dh_cnt



# import libraries
from datetime import timedelta
from typing import Any, List
# import locals
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event, EventSource
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.utils.Logger import Logger
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature

class TotalKelpTime(Feature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self.prev_time = timedelta(0)
        self.total_time = timedelta(0)
        self.idle_time = timedelta(0)
        self.on = False
        self.kelp = 'kelp'
### use count index like in region count, then run normal logic after that

    def Subfeatures(self) -> List[str]:
        return ["Seconds", "Active", "Active-Seconds", "Idle", "Idle-Seconds"]
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["begin_argument", "room_changed", "leave_argument", "complete_argument"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.app_version == 'Aqualab' or event.app_version == 'None':
            if event.EventSource == EventSource.GAME:
                if self.on:
                    self.total_time += (event.Timestamp - self.prev_time)
                    if (event.Timestamp - self.prev_time) > timedelta(minutes=1):
                        self.idle_time += (event.Timestamp - self.prev_time)
                    if not event.EventData.get('job_name', {}).get('string_value').startswith(self.kelp):
                        self.on = False
                else:
                    if event.EventData.get('job_name', {}).get('string_value').startswith(self.kelp):
                        self.on = True
                self.prev_time = event.Timestamp
        else:
            if event.EventSource == EventSource.GAME:
                if self.on:
                    self.total_time += (event.Timestamp - self.prev_time)
                    if (event.Timestamp - self.prev_time) > timedelta(minutes=1):
                        self.idle_time += (event.Timestamp - self.prev_time)
                    if not event.EventData.get('job_name', {}).get('string_value').startswith(self.kelp):
                        self.on = False
                else:
                    if event.EventData.get('job_name', {}).get('string_value').startswith(self.kelp):
                        self.on = True
                self.prev_time = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self.total_time is not None:
            return [self.total_time, self.total_time.total_seconds(), (self.total_time - self.idle_time), (self.total_time - self.idle_time).total_seconds(), self.idle_time, self.idle_time.total_seconds()]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***



import json
from typing import Dict, List, Optional

_enum_to_interactive_entry : Dict[int, str] = {
  0: 'tunic.entry_voteforgramps',
  1: 'tunic.entry_tunic',
  2: 'tunic.entry_cleanerslip',
  3: 'tunic.entry_basketballplaque',
  4: 'tunic.entry_teddytaken',
  5: 'tunic.entry_javajacket',
  6: 'tunic.entry_onthecase',
  7: 'tunic.entry_archivistcoffee',
  8: 'tunic.entry_expert',
  9: 'tunic.entry_logbook',
  10: 'tunic.entry_cleanercard',
  11: 'tunic.entry_microfiche',
  12: 'tunic.entry_newspaper',
  13: 'tunic.entry_taxidermy',
  14: 'tunic.entry_wellsid',
  15: 'tunic.entry_stacks_slip',
  16: 'tunic.entry_theodora',
  17: 'tunic.entry_scratches',
  18: 'tunic.entry_glasses',
  19: 'tunic.entry_teddyfree',
  20: 'tunic.entry_teddyfound',
  21: 'tunic.entry_key',
  22: 'tunic.entry_staffdir',
  23: 'tunic.entry_flag',
  24: 'tunic.entry_wellshoof',
  25: 'tunic.entry_tracks',
  26: 'tunic.entry_littering',
  27: 'tunic.entry_flaglady',
  28: 'tunic.entry_ecologyflag',
  29: 'tunic.entry_theta',
  30: 'tunic.entry_nelson',
  31: 'tunic.entry_activists'
}

def enum_to_interactive_entry(enum:int) -> str:
  return _enum_to_interactive_entry.get(enum, "INTERACTIVE ENTRY NOT FOUND!")

_interactive_entry_to_char = {v: chr(65+i+(i>25)*6) for i, v in _enum_to_interactive_entry.items()}
"""{'tunic.entry_voteforgramps': 'A',
 'tunic.entry_tunic': 'B',
 'tunic.entry_cleanerslip': 'C',
 'tunic.entry_basketballplaque': 'D',
 'tunic.entry_teddytaken': 'E',
 'tunic.entry_javajacket': 'F',
 'tunic.entry_onthecase': 'G',
 'tunic.entry_archivistcoffee': 'H',
 'tunic.entry_expert': 'I',
 'tunic.entry_logbook': 'J',
 'tunic.entry_cleanercard': 'K',
 'tunic.entry_microfiche': 'L',
 'tunic.entry_newspaper': 'M',
 'tunic.entry_taxidermy': 'N',
 'tunic.entry_wellsid': 'O',
 'tunic.entry_stacks_slip': 'P',
 'tunic.entry_theodora': 'Q',
 'tunic.entry_scratches': 'R',
 'tunic.entry_glasses': 'S',
 'tunic.entry_teddyfree': 'T',
 'tunic.entry_teddyfound': 'U',
 'tunic.entry_key': 'V',
 'tunic.entry_staffdir': 'W',
 'tunic.entry_flag': 'X',
 'tunic.entry_wellshoof': 'Y',
 'tunic.entry_tracks': 'Z',
 'tunic.entry_littering': 'a',
 'tunic.entry_flaglady': 'b',
 'tunic.entry_ecologyflag': 'c',
 'tunic.entry_theta': 'd',
 'tunic.entry_nelson': 'e',
 'tunic.entry_activists': 'f'}"""

def interactive_entry_to_char(entry_fqid) -> str:
  return _interactive_entry_to_char.get(entry_fqid) or ''

_assessment_answers : Dict[int, List[str]] = {
  1: ['tunic.entry_tunic','HACKME','tunic.entry_basketballplaque','tunic.entry_cleanerslip'],
  2: ['tunic.entry_cleanerslip','tunic.entry_expert','tunic.entry_cleanercard','tunic.entry_logbook',
     'tunic.entry_newspaper','tunic.entry_theodora'],
  3: ['tunic.entry_teddytaken','tunic.entry_javajacket','tunic.entry_archivistcoffee','tunic.entry_taxidermy'],
  4: ['tunic.entry_tracks', 'tunic.entry_ecologyflag', 'tunic.entry_theta','tunic.entry_nelson','tunic.entry_activists']
}

q_to_assessment : Dict[int, int] = {
 0: 1,
 1: 1,
 2: 1,
 3: 1,
 4: 2,
 5: 2,
 6: 2,
 7: 2,
 8: 2,
 9: 2,
 10: 3,
 11: 3,
 12: 3,
 13: 3,
 14: 4,
 15: 4,
 16: 4,
 17: 4,
 18: 4
}

assessement_to_lvl : Dict[int, int] = {
    1: 4,
    2: 12,
    3: 12,
    4: 22
}

assessment_to_last_q : Dict[int, int] = {
    1: 3,
    2: 9,
    3: 13,
    4: 18
}

_answer_fqid_to_q_num : Dict[str, int] = {
  'tunic.entry_tunic4': 0,
 'HACKME4': 1,
 'tunic.entry_basketballplaque4': 2,
 'tunic.entry_cleanerslip4': 3,
 'tunic.entry_cleanerslip12': 4,
 'tunic.entry_expert12': 5,
 'tunic.entry_cleanercard12': 6,
 'tunic.entry_logbook12': 7,
 'tunic.entry_newspaper12': 8,
 'tunic.entry_theodora12': 9,
 'tunic.entry_teddytaken12': 10,
 'tunic.entry_javajacket12': 11,
 'tunic.entry_archivistcoffee12': 12,
 'tunic.entry_taxidermy12': 13,
 'tunic.entry_tracks22': 14,
 'tunic.entry_ecologyflag22': 15,
 'tunic.entry_theta22': 16,
 'tunic.entry_nelson22': 17,
 'tunic.entry_activists22': 18
}

answer_to_question = lambda ans,lvl: _answer_fqid_to_q_num[f'{ans}{lvl}']

_answer_chars : List[Optional[str]] = ['B',
 None, # standin for D or C
 'D',
 'C',
 'C',
 'I',
 'K',
 'J',
 'M',
 'Q',
 'E',
 'F',
 'H',
 'N',
 'Z',
 'c',
 'd',
 'e',
 'f']

# objective is defined but what you have to click in order to start that objective.
# objective 0 is slightly different as you start the game with that objective
# do you always need an objective? idk.
fqid_to_enum : Dict[str, int] = {
  'tunic.historicalsociety.closet.gramps.intro_0_cs_0': 0,
  'tunic.historicalsociety.closet.teddy.intro_0_cs_0': 1,
  'tunic.historicalsociety.closet.notebook': 2,
  'tunic.historicalsociety.entry.groupconvo': 3,
  'tunic.historicalsociety.collection.cs': 4,
  'tunic.historicalsociety.collection.gramps.found': 5,
  'tunic.kohlcenter.halloffame.plaque.face.date': 6,
  'tunic.kohlcenter.halloffame.togrampa': 7,
  'tunic.capitol_0.hall.boss.chap1_finale_0': 8,
  'tunic.capitol_0.hall.boss.chap1_finale_1': 9,
  'tunic.capitol_0.hall.boss.chap1_finale_2': 10,
  'tunic.capitol_0.hall.chap1_finale_c': 11,
  'tunic.historicalsociety.closet_dirty.what_happened': 12,
  'tunic.historicalsociety.closet_dirty.trigger_scarf': 13,
  'tunic.historicalsociety.closet_dirty.trigger_coffee': 14,
  'tunic.historicalsociety.closet_dirty.gramps.news': 15,
  'tunic.historicalsociety.frontdesk.archivist.hello': 16,
  'tunic.historicalsociety.frontdesk.magnify': 17,
  'tunic.historicalsociety.frontdesk.archivist.have_glass': 18,
  'tunic.humanecology.frontdesk.worker.intro': 19,
  'tunic.humanecology.frontdesk.businesscards.card_bingo.bingo': 20,
  'tunic.drycleaner.frontdesk.worker.hub': 21,
  'tunic.drycleaner.frontdesk.logbook.page.bingo': 22,
  'tunic.drycleaner.frontdesk.worker.done': 23,
  'tunic.library.frontdesk.worker.hello': 24,
  'tunic.library.microfiche.reader.paper2.bingo': 25,
  'tunic.library.frontdesk.wellsbadge.hub': 26,
  'tunic.library.frontdesk.worker.wells': 27,
  'tunic.historicalsociety.frontdesk.archivist.newspaper': 28,
  'tunic.historicalsociety.stacks.journals.pic_2.bingo': 29,
  'tunic.capitol_1.hall.chap2_finale_c': 30,
  'tunic.capitol_1.hall.boss.chap2_finale_0': 31,
  'tunic.capitol_1.hall.boss.chap2_finale_1': 32,
  'tunic.capitol_1.hall.boss.chap2_finale_2': 33,
  'tunic.capitol_1.hall.boss.chap2_finale_3': 34,
  'tunic.capitol_1.hall.boss.chap2_finale_4': 35,
  'tunic.capitol_1.hall.boss.chap2_finale_5': 36,
  'tunic.capitol_1.hall.boss.chap2_finale_6': 37,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_0': 38,
  'tunic.capitol_1.hall.boss.chap2_teddy_finale_1': 39,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_2': 40,
  'tunic.capitol_1.hall.wells.chap2_teddy_finale_3': 41,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_4': 42,
  'tunic.historicalsociety.basement.ch3start': 43,
  'tunic.historicalsociety.basement.seescratches': 44,
  'tunic.historicalsociety.cage.teddy.trapped': 45,
  'tunic.historicalsociety.cage.glasses.afterteddy': 46,
  'tunic.historicalsociety.frontdesk.key': 47,
  'tunic.historicalsociety.cage.unlockdoor': 48,
  'tunic.historicalsociety.cage.confrontation': 49,
  'tunic.historicalsociety.basement.savedteddy': 50,
  'tunic.historicalsociety.collection_flag.gramps.flag': 51,
  'tunic.historicalsociety.entry.groupconvo_flag': 52,
  'tunic.historicalsociety.entry.wells.flag': 53,
  'tunic.historicalsociety.entry.boss.flag': 54,
  'tunic.wildlife.center.crane_ranger.crane': 55,
  'tunic.wildlife.center.remove_cup': 56,
  'tunic.wildlife.center.expert.removed_cup': 57,
  'tunic.wildlife.center.tracks.hub.deer': 58,
  'tunic.wildlife.center.coffee': 59,
  'tunic.wildlife.center.wells.animals': 60,
  'tunic.wildlife.center.wells.nodeer': 61,
  'tunic.flaghouse.entry.flag_girl.hello': 62,
  'tunic.flaghouse.entry.colorbook': 63,
  'tunic.flaghouse.entry.flag_girl.symbol': 64,
  'tunic.library.frontdesk.worker.flag': 65,
  'tunic.library.microfiche.reader_flag.paper2.bingo': 66,
  'tunic.library.frontdesk.worker.nelson': 67,
  'tunic.historicalsociety.frontdesk.archivist_glasses.confrontation': 68,
  'tunic.capitol_2.hall.chap4_finale_c': 69,
  'tunic.capitol_2.hall.boss.chap4_finale_0': 70,
  'tunic.capitol_2.hall.boss.chap4_finale_1': 71,
  'tunic.capitol_2.hall.boss.chap4_finale_2': 72,
  'tunic.capitol_2.hall.boss.chap4_finale_3': 73,
  'tunic.capitol_2.hall.boss.chap4_finale_4': 74,
  'tunic.capitol_2.hall.boss.chap4_finale_5': 75,
  'tunic.nelson.trail.conclusion': 76,
  'tunic.nelson.trail.coffee': 77,
  'tunic.nelson.trail.display': 78,
  'tunic.capitol_3.hall.chap5_finale_c': 79,
  'tunic.historicalsociety.closet.teddy.intro_0_cs_5': 80,
  'tunic.historicalsociety.closet.doorblock': 81,
  'tunic.historicalsociety.closet.photo': 82,
  'tunic.historicalsociety.closet.retirement_letter.hub': 83,
  'tunic.historicalsociety.basement.gramps.seeyalater': 84,
  'tunic.historicalsociety.basement.gramps.whatdo': 85,
  'tunic.historicalsociety.basement.janitor': 86,
  'tunic.historicalsociety.entry.block_tocollection': 87,
  'tunic.historicalsociety.entry.block_tomap1': 88,
  'tunic.historicalsociety.entry.block_tomap2': 89,
  'tunic.historicalsociety.entry.boss.flag_recap': 90,
  'tunic.historicalsociety.entry.boss.talktogramps': 91,
  'tunic.historicalsociety.entry.directory.closeup.archivist': 92,
  'tunic.historicalsociety.entry.gramps.hub': 93,
  'tunic.historicalsociety.entry.wells.flag_recap': 94,
  'tunic.historicalsociety.entry.wells.talktogramps': 95,
  'tunic.historicalsociety.collection': 96,
  'tunic.historicalsociety.collection.gramps.look_0': 97,
  'tunic.historicalsociety.collection.gramps.lost': 98,
  'tunic.historicalsociety.collection.slip': 99,
  'tunic.kohlcenter.halloffame.block_0': 100,
  'tunic.kohlcenter.halloffame.block_1': 101,
  'tunic.kohlcenter.halloffame.plaque.face': 102,
  'tunic.capitol_0.hall.boss.chap1_finale_0_fail': 103,
  'tunic.capitol_0.hall.boss.chap1_finale_1_fail': 104,
  'tunic.capitol_0.hall.boss.chap1_finale_plaquefirst_0': 105,
  'tunic.capitol_0.hall.boss.chap1_finale_plaquefirst_0_fail': 106,
  'tunic.capitol_0.hall.boss.chap1_finale_plaquefirst_1': 107,
  'tunic.capitol_0.hall.boss.chap1_finale_slipfirst_0': 108,
  'tunic.capitol_0.hall.boss.chap1_finale_slipfirst_0_fail': 109,
  'tunic.capitol_0.hall.boss.chap1_finale_slipfirst_1': 110,
  'tunic.capitol_0.hall.boss.talktogramps': 111,
  'tunic.capitol_0.hall.gramps.hub': 112,
  'tunic.historicalsociety.closet_dirty.door_block_clean': 113,
  'tunic.historicalsociety.closet_dirty.door_block_talk': 114,
  'tunic.historicalsociety.closet_dirty.gramps.archivist': 115,
  'tunic.historicalsociety.closet_dirty.gramps.helpclean': 116,
  'tunic.historicalsociety.closet_dirty.gramps.nothing': 117,
  'tunic.historicalsociety.closet_dirty.photo': 118,
  'tunic.historicalsociety.stacks.block': 119,
  'tunic.historicalsociety.stacks.journals': 120,
  'tunic.historicalsociety.stacks.journals_flag': 121,
  'tunic.historicalsociety.stacks.journals_flag.pic_0.bingo': 122,
  'tunic.historicalsociety.stacks.journals_flag.pic_1.bingo': 123,
  'tunic.historicalsociety.stacks.journals_flag.pic_2.bingo': 124,
  'tunic.historicalsociety.stacks.outtolunch': 125,
  'tunic.historicalsociety.frontdesk.archivist.foundtheodora': 126,
  'tunic.historicalsociety.frontdesk.archivist.have_glass_recap': 127,
  'tunic.historicalsociety.frontdesk.archivist.need_glass_0': 128,
  'tunic.historicalsociety.frontdesk.archivist.need_glass_1': 129,
  'tunic.historicalsociety.frontdesk.archivist.newspaper_recap': 130,
  'tunic.historicalsociety.frontdesk.archivist_glasses.confrontation_recap': 131,
  'tunic.historicalsociety.frontdesk.block_magnify': 132,
  'tunic.humanecology.frontdesk.block_0': 133,
  'tunic.humanecology.frontdesk.block_1': 134,
  'tunic.humanecology.frontdesk.businesscards': 135,
  'tunic.humanecology.frontdesk.worker.badger': 136,
  'tunic.drycleaner.frontdesk.block_0': 137,
  'tunic.drycleaner.frontdesk.block_1': 138,
  'tunic.drycleaner.frontdesk.logbook': 139,
  'tunic.drycleaner.frontdesk.worker.done2': 140,
  'tunic.drycleaner.frontdesk.worker.takealook': 141,
  'tunic.library.frontdesk.block_badge': 142,
  'tunic.library.frontdesk.block_badge_2': 143,
  'tunic.library.frontdesk.block_nelson': 144,
  'tunic.library.frontdesk.worker.droppedbadge': 145,
  'tunic.library.frontdesk.worker.flag_recap': 146,
  'tunic.library.frontdesk.worker.hello_short': 147,
  'tunic.library.frontdesk.worker.nelson_recap': 148,
  'tunic.library.frontdesk.worker.preflag': 149,
  'tunic.library.frontdesk.worker.wells_recap': 150,
  'tunic.library.microfiche.block_0': 151,
  'tunic.library.microfiche.reader': 152,
  'tunic.capitol_1.hall.boss.chap2_finale_0_fail': 153,
  'tunic.capitol_1.hall.boss.chap2_finale_1_fail': 154,
  'tunic.capitol_1.hall.boss.chap2_finale_2_fail': 155,
  'tunic.capitol_1.hall.boss.chap2_finale_3_fail': 156,
  'tunic.capitol_1.hall.boss.chap2_finale_4_fail': 157,
  'tunic.capitol_1.hall.boss.chap2_finale_5_fail': 158,
  'tunic.capitol_1.hall.boss.chap2_finale_6_fail': 159,
  'tunic.capitol_1.hall.boss.chap2_teddy_finale_1_fail': 160,
  'tunic.capitol_1.hall.boss.haveyougotit': 161,
  'tunic.capitol_1.hall.boss.writeitup': 162,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_0_fail': 163,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_2_fail': 164,
  'tunic.capitol_1.hall.gramps.chap2_teddy_finale_4_fail': 165,
  'tunic.capitol_1.hall.wells.chap2_teddy_finale_3_fail': 166,
  'tunic.capitol_1.hall.wells.hub': 167,
  'tunic.historicalsociety.cage.glasses.beforeteddy': 168,
  'tunic.historicalsociety.cage.lockeddoor': 169,
  'tunic.historicalsociety.cage.need_glasses': 170,
  'tunic.historicalsociety.collection_flag.gramps.recap': 171,
  'tunic.wildlife.center.expert.recap': 172,
  'tunic.wildlife.center.fox.concern': 173,
  'tunic.wildlife.center.wells.animals2': 174,
  'tunic.wildlife.center.wells.nodeer_recap': 175,
  'tunic.flaghouse.entry.flag_girl.hello_recap': 176,
  'tunic.flaghouse.entry.flag_girl.symbol_recap': 177,
  'tunic.capitol_2.hall.boss.chap4_finale_0_fail': 178,
  'tunic.capitol_2.hall.boss.chap4_finale_1_fail': 179,
  'tunic.capitol_2.hall.boss.chap4_finale_2_fail': 180,
  'tunic.capitol_2.hall.boss.chap4_finale_3_fail': 181,
  'tunic.capitol_2.hall.boss.chap4_finale_4_fail': 182,
  'tunic.capitol_2.hall.boss.chap4_finale_5_fail': 183,
  'tunic.capitol_2.hall.boss.haveyougotit': 184,
  'tunic.nelson.trail.gramps.question': 185,
  'tunic.capitol_3.hall.crowd.hub': 186,
  'tunic.capitol_3.hall.gramps.hub': 187,
  'tunic.capitol_3.hall.teddy.speak': 188}

max_objective : int = 79

save_code_to_obj : Dict[str, int] = {
  'startgame': 0,
  'notebook': 3,
  'wiscwonders': 4,
  'mysteryslip': 6,
  'plaque': 8,
  'notajersey': 12,
  'trashed': 13,
  'archivist': 19,
  'textile': 21,
  'logbook': 23,
  'suffragist': 26,
  'taxidermist': 28,
  'wellsdidit': 30,
  'saveteddy': 43,
  'scratches': 45,
  'hesalive': 46,
  'akey': 48,
  'rescued': 50,
  'backtowork': 52,
  'sadanimals': 62,
  'flaglady': 65,
  'ecologists': 67,
  'donethework': 69,
  'sunset': 77
}

# level to first objective in level.
level_to_start_obj : Dict[int, int] = {
  0: 0, 1: 3, 2: 4, 3: 6, 4: 8, 5: 12, 6: 13, 7: 19, 8: 21, 9: 23, 10: 26, 11: 28, 12: 30, 13: 43,
  14: 45, 15: 46, 16: 48, 17: 50, 18: 52, 19: 62, 20: 65, 21: 67, 22: 69, 23: 77
}

def quiz_question_to_index(quiz_num:int, question_num:int) -> int:
  """See: https://github.com/fielddaylab/jo_wilder/blob/master/src/scenes/quiz.js

  Each quiz is defined as a seperate dictionary, which means the index of each
  question is specific to its position in that dictionary, rather than its position
  in the overall question list.

  The quiz_indexes dictionary contains the proper indexing for all questions in each 
  of the 5 quizzes.

  Returns:
    The remapped index given a quiz and question number.
  """
  quiz_indexes : Dict[int, Dict[int, int]] = {
    0: {0: 0, 1: 1},
    2: {0: 2, 1: 3, 2: 4, 3: 5},
    3: {0: 6, 1: 7, 2: 8, 3: 9},
    4: {0: 10, 1: 11, 2: 12, 3: 13},
    5: {0: 14, 1: 15, 2: 16, 3: 17}
  }

  return quiz_indexes[quiz_num][question_num]


def quizn_to_index(quizn:int) -> int:
  """See: https://github.com/fielddaylab/jo_wilder/blob/master/src/scenes/quiz.js

  For some reason there are 5 quizzes, but there is no quiz numbered 1.
  
  Returns:
    The correct quiz number for quizzes 2-5, or 0 for quiz 0.
  """
  return quizn - 1 if quizn >= 2 else quizn

if __name__ == '__main__':
 print({i:v for i,(k,v) in enumerate(save_code_to_obj.items())})


## import standard libraries
import abc
from enum import IntEnum
from typing import Dict, List, Tuple, Optional

# import local files
from coding.Code import Code
from coding.Coder import Coder
from ogd.core.interfaces.Interface import Interface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.utils.Logger import Logger

class CodingInterface(Interface):

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _allCoders(self) -> Optional[List[Coder]]:
        pass

    @abc.abstractmethod
    def _createCoder(self, coder_name:str) -> bool:
        pass

    @abc.abstractmethod
    def _getCodeWordsByGame(self, game_id:str) -> Optional[List[str]]:
        pass

    @abc.abstractmethod
    def _getCodeWordsByCoder(self, coder_id:str) -> Optional[List[str]]:
        pass

    @abc.abstractmethod
    def _getCodeWordsBySession(self, session_id:str) -> Optional[List[str]]:
        pass

    @abc.abstractmethod
    def _getCodesByGame(self, game_id:str) -> Optional[List[Code]]:
        pass

    @abc.abstractmethod
    def _getCodesByCoder(self, coder_id:str) -> Optional[List[Code]]:
        pass

    @abc.abstractmethod
    def _getCodesBySession(self, session_id:str) -> Optional[List[Code]]:
        pass

    @abc.abstractmethod
    def _createCode(self, code:str, coder_id:str, events:List[Code.EventID], notes:Optional[str]=None) -> bool:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self):
        super().__init__()

    def __del__(self):
        self.Close()

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def AllCoders(self) -> Optional[List[Coder]]:
        ret_val = None
        if self.IsOpen():
            ret_val = self._allCoders()
        else:
            Logger.Log("Can't retrieve list of all Coders, the source interface is not open!")
        return ret_val

    def CreateCoder(self, coder_name:str) -> bool:
        ret_val = False
        if self.IsOpen():
            ret_val = self._createCoder(coder_name=coder_name)
        else:
            Logger.Log("Can't create Coder, the source interface is not open!")
        return ret_val

    def GetCodes(self, mode:IDMode, id:str):
        match mode:
            case IDMode.GAME:
                self._getCodesByGame(game_id=id)
            case IDMode.USER:
                self._getCodesByCoder(coder_id=id)
            case IDMode.SESSION:
                self._getCodesBySession(session_id=id)
            case _:
                raise NotImplementedError(f"The given retrieval mode '{mode}' is not supported for retrieving codes!")

    def GetCodeWords(self, mode:IDMode, id:str):
        match mode:
            case IDMode.GAME:
                self._getCodeWordsByGame(game_id=id)
            case IDMode.USER:
                self._getCodeWordsByCoder(coder_id=id)
            case IDMode.SESSION:
                self._getCodeWordsBySession(session_id=id)
            case _:
                raise NotImplementedError(f"The given retrieval mode '{mode}' is not supported for retrieving code words!")

    def CreateCode(self, code:str, coder_id:str, events:List[Code.EventID], notes:Optional[str]=None) -> bool:
        ret_val = False
        if self.IsOpen():
            ret_val = self._createCode(code=code, coder_id=coder_id, events=events, notes=notes)
        else:
            Logger.Log("Can't create Code, the source interface is not open!")
        return ret_val

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



from ogd.core.models import Event
import typing
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class TotalFails(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._fail_count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["FAIL.0"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._fail_count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._fail_count]

    # *** Optionally override public functions. ***



# import libraries
import json
import logging
from collections import defaultdict
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class ActiveJobs(Feature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        self._job_map = job_map
        super().__init__(params=params)
        self._current_user_code = None
        self._last_started_id = None
        self._active_jobs = defaultdict(list)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["accept_job", "switch_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        _current_job = event.GameState.get('job_name', event.EventData.get('job_name', None))
        if self._validate_job(_current_job):
            user_code = event.UserID

            if (self._current_user_code is not None) and (self._current_user_code != user_code) and (self._current_user_code not in self._active_jobs[self._last_started_id]):
                # if we found a new user, then previous user must have left off on whatever their active job was.
                # so, add the user to the list for that job
                self._active_jobs[self._last_started_id].append(self._current_user_code)
            self._current_user_code = user_code # in either case, set latest user as "current"
            self._last_started_id = _current_job # In either case, set latest job name as "current".

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        ret_val = self._active_jobs

        if (self._last_started_id is not None) and (self._current_user_code not in self._active_jobs[self._last_started_id]):
            ret_val[self._last_started_id].append(self._current_user_code) # whatever last event was, assume player left off there.

        return [json.dumps(ret_val)]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    # *** Other local functions
    def _validate_job(self, job_data):
        ret_val : bool = False
        if job_data and job_data in self._job_map:
            ret_val = True
        else:
            self.WarningMessage(f"Got invalid job_name data in JobsAttempted")
        return ret_val



import json
import logging
import os
from datetime import datetime
from typing import Dict, Final, List, Tuple, Optional
# import locals
from ogd.core.interfaces.BigQueryInterface import BigQueryInterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.utils.Logger import Logger

AQUALAB_MIN_VERSION : Final[float] = 6.2

class BQFirebaseInterface(BigQueryInterface):
    """Subclass of the general BigQueryInterface, with re-implemented queries to handle old Firebase unnest bullshit.
    """

    # *** BUILT-INS ***

    def __init__(self, game_id:str, config:GameSourceSchema, fail_fast:bool):
        super().__init__(game_id=game_id, config=config, fail_fast=fail_fast)
        self.Open()

    # *** RE-IMPLEMENT ABSTRACT FUNCTIONS ***

    def _allIDs(self) -> List[str]:
        query = f"""
            SELECT DISTINCT param.value.int_value AS session_id
            FROM `{self.DBPath()}`,
            UNNEST(event_params) AS param
            WHERE param.key = "ga_session_id"
        """
        Logger.Log(f"BQ-Firebase: Running query for all ids:\n{query}", logging.DEBUG, depth=3)
        data = self._client.query(query)
        ids = [str(row['session_id']) for row in data]
        return ids if ids != None else []

    def _fullDateRange(self) -> Dict[str, datetime]:
        query = f"""
            WITH datetable AS
            (
                SELECT event_date, event_timestamp,
                FORMAT_DATE('%m-%d-%Y', PARSE_DATE('%Y%m%d', event_date)) AS date, 
                FORMAT_TIME('%T', TIME(TIMESTAMP_MICROS(event_timestamp))) AS time,
                FROM `{self.DBPath()}`
            )
            SELECT MIN(concat(date, ' ', time)), MAX(concat(date, ' ', time))
            FROM datetable
        """
        Logger.Log(f"BQ-Firebase: Running query for full date range:\n{query}", logging.DEBUG, depth=3)
        data = list(self._client.query(query))
        return {'min':data[0][0], 'max':data[0][1]}

    def _rowsFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None, exclude_rows:Optional[List[str]]=None) -> List[Tuple]:
        events = None
        if self._client != None:
            query = self._generateRowFromIDQuery(id_list=id_list, id_mode=id_mode, exclude_rows=exclude_rows)
            Logger.Log(f"BQ-Firebase: Running query for rows from IDs:\n{query}", logging.DEBUG, depth=3)
            data = self._client.query(query)
            events = []
            for row in data:
                items = tuple(row.items())
                event = []
                for item in items:
                    match item[0]:
                        case "event_params":
                            _params = {param['key']:param['value'] for param in item[1]}
                            event.append(json.dumps(_params, sort_keys=True))
                        case "device":
                            event.append(json.dumps(item[1], sort_keys=True))
                        case _:
                            event.append(item[1])
                events.append(tuple(event))
        return events if events != None else []

    def _IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]] = None) -> List[str]:
        ret_val = []
        str_min, str_max = min.strftime("%Y%m%d"), max.strftime("%Y%m%d")
        query = f"""
            SELECT DISTINCT param.value.int_value AS session_id
            FROM `{self.DBPath(min_date=min.date(), max_date=max.date())}`,
            UNNEST(event_params) AS param
            WHERE param.key = "ga_session_id"
            AND _TABLE_SUFFIX BETWEEN '{str_min}' AND '{str_max}'
        """
        Logger.Log(f"BQ-Firebase: Running query for ids from dates:\n{query}", logging.DEBUG, depth=3)
        data = self._client.query(query)
        ids = [str(row['session_id']) for row in data]
        if ids is not None:
            ret_val = ids
        return ret_val

    def _datesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None) -> Dict[str, datetime]:
        match id_mode:
            case IDMode.SESSION:
                id_string = ','.join([f"{x}" for x in id_list])
                where_clause = f"""
                    WHERE param.key = "ga_session_id"
                    AND param.value.int_value IN ({id_string})
                """
            case IDMode.USER:
                id_string = ','.join([f"'{x}'" for x in id_list])
                where_clause = f"""
                    WHERE param.key = "user_code"
                    AND param.value.string_value IN ({id_string})
                """
            case _:
                Logger.Log(f"Invalid ID mode given (name={id_mode.name}, val={id_mode.value}), defaulting to session mode.", logging.WARNING, depth=3)
                id_string = ','.join([f"{x}" for x in id_list])
                where_clause = f"""
                    WHERE param.key = "ga_session_id"
                    AND param.value.int_value IN ({id_string})
                """
        query = f"""
            WITH datetable AS
            (
                SELECT event_date, event_timestamp, event_params,
                FORMAT_DATE('%m-%d-%Y', PARSE_DATE('%Y%m%d', event_date)) AS date, 
                FORMAT_TIME('%T', TIME(TIMESTAMP_MICROS(event_timestamp))) AS time,
                FROM `{self.DBPath()}`
            )
            SELECT MIN(concat(date, ' ', time)), MAX(concat(date, ' ', time))
            FROM datetable,
            UNNEST(event_params) AS param
            {where_clause}
        """
        Logger.Log(f"BQ-Firebase: Running query for dates from IDs:\n{query}", logging.DEBUG, depth=3)
        data = list(self._client.query(query))
        ret_val : Dict[str, datetime] = {}
        if len(data) == 1:
            dates = data[0]
            if len(dates) == 2 and dates[0] is not None and dates[1] is not None:
                ret_val = {'min':datetime.strptime(dates[0], "%m-%d-%Y %H:%M:%S"), 'max':datetime.strptime(dates[1], "%m-%d-%Y %H:%M:%S")}
            else:
                Logger.Log(f"BQFirebaseInterface query did not give both a min and a max, setting both to 'now'", logging.WARNING, depth=3)
                ret_val = {'min':datetime.now(), 'max':datetime.now()}
        else:
            Logger.Log(f"BQ-Firebase: Query did not return any results, setting both min and max to 'now'", logging.WARNING, depth=3)
            ret_val = {'min':datetime.now(), 'max':datetime.now()}
        return ret_val

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _generateRowFromIDQuery(self, id_list:List[str], id_mode:IDMode, exclude_rows:Optional[List[str]]=None) -> str:
    # 2) Set up clauses to select based on Session ID or Player ID.
        session_clause : str = ""
        player_clause  : str = ""
        match id_mode:
            case IDMode.SESSION:
                id_string = ','.join([f"{x}" for x in id_list])
                session_clause = f"param_session.key = 'ga_session_id' AND param_session.value.int_value IN ({id_string})"
                player_clause  = f"(param_user.key   = 'user_code'     OR  param_user.key = 'undefined')"
            case IDMode.USER:
                id_string = ','.join([f"'{x}'" for x in id_list])
                session_clause = f"param_session.key = 'ga_session_id'"
                player_clause  = f"(param_user.key   = 'user_code' OR param_user.key = 'undefined') AND param_user.value.string_value IN ({id_string})"
            case _:
                Logger.Log(f"BQ-Firebase: Invalid ID mode given (name={id_mode.name}, val={id_mode.value}), defaulting to session mode.", logging.WARNING, depth=3)
                id_string = ','.join([f"{x}" for x in id_list])
                session_clause = f"param_session.key = 'ga_session_id' AND param_session.value.int_value IN ({id_string})"
                player_clause  = f"(param_user.key   = 'user_code' OR param_user.key = 'undefined')"
    # 3) Set up WHERE clause based on whether we need Aqualab min version or not.
        match self._game_id:
            case "SHADOWSPECT" | "SHIPWRECKS":
                where_clause = \
                f"WHERE {session_clause}"
            case _:
                where_clause = \
                f"""WHERE param_app_version.key = 'app_version'
                    AND   param_log_version.key = 'log_version'
                    AND   {session_clause}
                    AND   {player_clause}"""
    # 4) Set up actual query
        query = ""
        match self._game_id:
            case "SHADOWSPECT" | "SHIPWRECKS":
                query = f"""
                    SELECT event_name, event_params, device, platform,
                    concat(FORMAT_DATE('%Y-%m-%d', PARSE_DATE('%Y%m%d', event_date)), FORMAT_TIME('T%H:%M:%S.00', TIME(TIMESTAMP_MICROS(event_timestamp)))) AS timestamp,
                    null as app_version,
                    null as log_version,
                    param_session.value.int_value as session_id,
                    null as fd_user_id
                    FROM `{self.DBPath()}`
                    CROSS JOIN UNNEST(event_params) AS param_session
                    {where_clause}
                    ORDER BY `session_id`, `timestamp` ASC;
                """
            case _:
                # TODO Order by user_id, and by timestamp within that.
                # Note that this could prove to be wonky when we have more games without user ids,
                # will need to really rethink this when we start using new system.
                # Still, not a huge deal because most of these will be rewritten at that time anyway.
                query = f"""
                    SELECT event_name, event_params, device, platform,
                    concat(FORMAT_DATE('%Y-%m-%d', PARSE_DATE('%Y%m%d', event_date)), FORMAT_TIME('T%H:%M:%S.00', TIME(TIMESTAMP_MICROS(event_timestamp)))) AS timestamp,
                    param_app_version.value.string_value as app_version,
                    param_log_version.value.int_value as log_version,
                    param_session.value.int_value as session_id,
                    param_user.value.string_value as fd_user_id
                    FROM `{self.DBPath()}`
                    CROSS JOIN UNNEST(event_params) AS param_app_version
                    CROSS JOIN UNNEST(event_params) AS param_log_version
                    CROSS JOIN UNNEST(event_params) AS param_session
                    CROSS JOIN UNNEST(event_params) AS param_user
                    {where_clause}
                    ORDER BY `fd_user_id`, `timestamp` ASC;
                """
        return query



# import standard libraries
from datetime import datetime, timedelta
from typing import Any, Callable, List, Optional
# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

class DiveSiteNoEvidence(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    # *** Implement abstract functions ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"]

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if event.EventName == "scene_changed":
            # as soon as scene changes, reset state.
            self._has_triggered = False
            self._time_since_evidence = 0
            scene : str = event.EventData['scene_name']
            if scene.startswith("RS-"):
                self._dive_site = scene
                self._in_dive = True
                # if we just started a dive, use now as starting point for counting time without evidence.
                self._last_evidence_time = event.Timestamp
            else:
                # if we changed to a different scene, reset state.
                self._dive_site = None
                self._in_dive = False
                self._last_evidence_time = None
        elif event.EventName == "receive_entity":
            self._last_evidence_time = event.Timestamp

        if self._in_dive and self._last_evidence_time is not None and not self._has_triggered:
            self._time_since_evidence = (event.Timestamp - self._last_evidence_time).total_seconds()
        self._current_job = event.GameState.get('job_name', event.EventData.get('job_name'))
        return
    
    def _trigger_condition(self) -> bool:
        """_summary_
        """
        if self._in_dive and self._time_since_evidence > self._threshold and not self._has_triggered:
            self._has_triggered = True
            return True
        else:
            return False

    def _trigger_event(self) -> Optional[Event]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : Event = self.GenerateEvent(app_id="AQUALAB",
                                             event_name="DiveSiteNoEvidence", event_data={"dive_site":self._dive_site},
                                             game_state={"job_name":self._current_job})
        return ret_val

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None], threshold:float):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._in_dive : bool    = False
        self._dive_site = None
        self._threshold : float = threshold
        self._last_evidence_time  : Optional[datetime] = None
        self._time_since_evidence : float = 0
        self._has_triggered : bool = False
        self._current_job = None

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import json
from typing import Any, List, Optional
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature




class TotalFails(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._fail_count : int = 0;
        # >>> create/initialize any variables to track feature extractor state <<<
        #
        # e.g. To track whether extractor found a click event yet:
        # self._found_click : bool = False

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["level_fail"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
        
        self._fail_count +=1
        
        
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        
        # >>> use state variables to calculate the return value(s) of the base Feature and any Subfeatures. <<<
        # >>> put the calculated value(s) into a list as the function return value. <<<
        # >>> definitely don't return ["template"], unless you really find that useful... <<<
        #
        # e.g. use the self._found_click, which was created/initialized in __init__(...), and updated in _updateFromEvent(...):
        # if self._found_click:
        #     ret_val = [True]
        # else:
        #     ret_val = [False]
        # return ret_val
        #
        # note the code above is redundant, we could just return [self._found_click] to get the same result;
        # the more-verbose code is here for illustrative purposes.
        
        return [self._fail_count]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return []
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



## import standard libraries
import abc
import logging
from typing import Dict, List, Type, Optional, Set
# import locals
from ogd.core.registries.GeneratorRegistry import GeneratorRegistry
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.processors.Processor import Processor
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

## @class Processor
class GeneratorProcessor(Processor):

    # *** ABSTRACTS ***

    ## Abstract declaration of a function to get the calculated value of the feature, given data seen so far.

    @property
    @abc.abstractmethod
    def _mode(self) -> ExtractionMode:
        pass

    ## Abstract declaration of a function to get the names of all features.
    @abc.abstractmethod
    def _getGeneratorNames(self) -> List[str]:
        pass

    @property
    @abc.abstractmethod
    def _playerID(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def _sessionID(self) -> str:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_schema: GameSchema, LoaderClass:Type[GeneratorLoader], feature_overrides:Optional[List[str]]=None):
        super().__init__(game_schema=game_schema)
        self._overrides   : Optional[List[str]]   = feature_overrides
        self._loader      : GeneratorLoader       = LoaderClass(player_id=self._playerID, session_id=self._sessionID, game_schema=self._game_schema,
                                                                mode=self._mode, feature_overrides=self._overrides)
        self._registry    : Optional[GeneratorRegistry] = None # Set to 0, let subclasses create own instances.

    def __str__(self):
        return f""

    @property
    def GeneratorNames(self) -> List[str]:
        # TODO: add error handling code, if applicable.
        return self._getGeneratorNames()

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def ProcessFeatureData(self, feature_list:List[FeatureData]) -> None:
        if self._registry is not None:
            for feature in feature_list:
                self._registry.UpdateFromFeatureData(feature=feature)
        else:
            Logger.Log(f"Processor has no registry, skipping FeatureData.", logging.WARN)

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import necessary libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class FailCount(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.fail_counts: Dict[str, int] = {
            "CityFailed": 0,
            "TooManyBlooms": 0,
            "OutOfMoney": 0
        }

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["lose_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        lose_condition = event.EventData.get("lose_condition", "")
        if lose_condition:
            if lose_condition in self.fail_counts:
                self.fail_counts[lose_condition] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        _total = sum([val for val in self.fail_counts.values()])
        return [_total, self.fail_counts]

    def Subfeatures(self) -> List[str]:
        return ["Breakdown"]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



## @namespace utils
#  A module of utility functions used in the feature_extraction_to_csv project
import json
import logging
import os
from importlib.resources import files
from pathlib import Path
from typing import Any, Dict, Optional, List
# import locals
from ogd.core.utils.Logger import Logger

map = Dict[str, Any]
ExportRow = List[Any]

## Function to open a given JSON file, and retrieve the data as a Python object.
def loadJSONFile(filename:str, path:Path = Path("./"), search_in_src:bool = False, autocorrect_extension:bool = True) -> Dict[str, Any]:
    """Function to open a given JSON file, and retrieve the data as a Python object.

    :param filename: The name of the JSON file. If the file extension is not .json, then ".json" will be appended.
    :type filename: str
    :param path: The path (relative or absolute) to the folder containing the JSON file. Defaults to Path("./")
    :type path: Path, optional
    :param search_in_src: When True, prepends "src" to the path when searching for file directly, defaults to False
    :type search_in_src: bool, optional
    :param autocorrect_extension: When False, overrides default behavior and will not append .json to a filename with other extension, defaults to True
    :type autocorrect_extension: bool, optional
    :raises err: _description_
    :return: A python object parsed from the JSON.
    :rtype: Dict[Any, Any]
    """
    if autocorrect_extension and not filename.lower().endswith(".json"):
        Logger.Log(f"Got a filename that didn't end with .json: {filename}, appending .json", logging.DEBUG)
        filename = filename + ".json"
    # once we've validated inputs, try actual loading and reading.
    file_path = Path("src") / path / filename if search_in_src else path / filename
    try:
        with open(file_path, "r") as json_file:
            return json.loads(json_file.read())
    except FileNotFoundError as err:
        Logger.Log(f"Could not load JSON file, {file_path} could not be found from {os.getcwd()}, trying to find within package.", logging.WARNING)
        package_file_path = None
        try:
            package_file_path = files(".".join(path.parts)).joinpath(filename)
            with package_file_path.open() as json_file:
                return json.loads(json_file.read())
        except ModuleNotFoundError as err:
            Logger.Log(f"Could not load JSON file, unable to search in module for {path}, got the following error:\n{err.msg}.", logging.WARNING)
            raise err
        except FileNotFoundError as err:
            Logger.Log(f"Could not load JSON file from package, {package_file_path} does not exist.", logging.WARNING)
            raise err



# import standard libraries
from datetime import datetime, timedelta
from time import time
from typing import Callable, Final, List, Optional, Union
# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

class Idle(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    DEFAULT_IDLE_LEVEL : Final[int] = 30


    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None], idle_level:Optional[int]):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._found = False
        self._sess_id = "Unknown"
        self._last_action_time = None
        self._job_name = "Unknown"
        self._idle_level = 0
        self._idle_time: Optional[Union[float, timedelta]] = None
        if idle_level is not None:
            self._idle_threads: timedelta = timedelta(seconds=idle_level)
        else:
            self._idle_threads: timedelta = timedelta(seconds=Idle.DEFAULT_IDLE_LEVEL)

    # *** Implement abstract functions ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"] # >>> fill in names of events this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if event.EventName == "script_fired" and ("somethingLeft" or "nothingLeft") in event.EventData.get("node_id", ""):
            return
        if self._sess_id == "Unknown":
            self._sess_id = event.SessionID
        elif self._sess_id != event.SessionID:
            self._sess_id = event.SessionID
            self._last_action_time = None
        if not self._last_action_time:
            self._last_action_time = event.Timestamp
            return

        self._idle_time = event.Timestamp - self._last_action_time
        self._last_action_time = event.Timestamp
        if self._idle_time > self._idle_threads:
            self._found = True
            self._idle_time = self._idle_time / timedelta(seconds=1)
            self._idle_level = self._idle_threads / timedelta(seconds=1)
            self._sess_id = event.SessionID
            self._job_name = event.GameState.get('job_name', event.EventData.get('job_name', "JOB NAME NOT FOUND"))
        return

    def _trigger_condition(self) -> bool:
        if self._found:
            self._found = False
            return True
        else:
            return False

    def _trigger_event(self) -> DetectorEvent:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        _event_data = {
            "level" : self._idle_level,
            "time": self._idle_time,
            "job_name": self._job_name
        }
        ret_val : DetectorEvent = self.GenerateEvent(session_id=self._sess_id, app_id="AQUALAB",
                                                     event_name="Idle", event_data=_event_data)
        return ret_val



## @package DataToCSV.py
#  A package to handle processing of stuff from our database,
#  for export to CSV files.

## import standard libraries
import logging
import math
import subprocess
import traceback
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Tuple, Type, Optional

## import local files
from ogd import games
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.managers.EventManager import EventManager
from ogd.core.managers.FeatureManager import FeatureManager
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.schemas.configs.ConfigSchema import ConfigSchema
from ogd.core.requests.Request import Request
from ogd.core.requests.RequestResult import RequestResult
from ogd.core.utils.Logger import Logger

Slice = List[str]

## @class ExportManager
#  A class to export features and raw data, given a Request object.
class ExportManager:
    """ExportManager class.
    Use this class to carry out a request for a data export, by passing along an instance of the `Request` class to the ExecuteRequest function.
    """

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, config:ConfigSchema):
        """Constructor for an ExportManager object.
        Simply sets the config for the manager. All other data comes from a request given to the manager.

        :param settings: [description]
        :type settings: [type]
        """
        self._config      : ConfigSchema = config
        self._event_mgr   : Optional[EventManager]   = None
        self._feat_mgr    : Optional[FeatureManager] = None
        self._debug_count : int                      = 0

    def __str__(self):
        return f"ExportManager"

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def ExecuteRequest(self, request:Request) -> RequestResult:
        """Carry out the export given by a request.
        Each request has a game id, an interface for getting the data, a data range, the output type(s),
        the locations for output (to file or return value),
        and an optional list of features to override the configured features for a game.

        :param request: [description]
        :type request: Request
        :return: [description]
        :rtype: Dict[str,Any]
        """
        ret_val : RequestResult = RequestResult(msg="No Export")

        Logger.Log(f"Executing request: {str(request)}", logging.INFO)
        start = datetime.now()
        try:
        # 1. Pre-processing
            Logger.Log(f"Setting up file, event, and feature managers as pre-processing...", logging.INFO)
            self._preProcess(request=request)
            Logger.Log(f"Done", logging.INFO)
            Logger.Log(f"Executing...", logging.INFO)
            _sess_ids : List[str] = request.RetrieveIDs() or []
            for outerface in request.Outerfaces:
                outerface.SessionCount = len(_sess_ids)

        # 2. Process slices
            Logger.Log(f"Preparing to process {len(_sess_ids)} sessions...", logging.INFO, depth=1)
            self._processSlices(request=request, ids=_sess_ids)
            Logger.Log(f"Done", logging.INFO, depth=1)

        # 3. Output population/player features as post-slicing data.
            Logger.Log(f"Outputting post-process data...", logging.INFO, depth=2)
            self._postProcess(request=request)
            Logger.Log(f"Done", logging.INFO)

            ret_val.SessionCount = len(_sess_ids)
            ret_val.RequestSucceeded(msg=f"Successfully executed data request {request}.")
        except ValueError as err:
            msg = f"Failed to execute data request {str(request)}, an invalid value was found:\n{str(err)}\n{traceback.format_exc()}"
            ret_val.RequestErrored(msg=msg)
        except Exception as err:
            msg = f"Failed to execute data request {str(request)}, an unexpected error occurred:\n{type(err)} {str(err)}\n{traceback.format_exc()}"
            ret_val.RequestErrored(msg=msg)
        finally:
            time_delta = datetime.now() - start
            ret_val.Duration = time_delta
            return ret_val

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _receiveEventTrigger(self, event:Event) -> None:
        # TODO: consider how to put a limit on times this runs, based on how big export is.
        if self._debug_count < 5:
            Logger.Log(f"{self} received an event trigger.", logging.DEBUG)
            self._debug_count += 1
        self._processEvent(next_event=event)

    def _preProcess(self, request:Request) -> None:
        """
        Pre-processing of request, setting up required components before we retrieve session data.

        In particular, this function:
        1. Sets up loader class
        2. Sets up event manager
        3. Sets up feature manager
        4. Opens request outerfaces
        :param request: The export request being processed
        :type request: Request
        """
        _games_path  = Path(games.__file__) if Path(games.__file__).is_dir() else Path(games.__file__).parent
        _game_schema  : GameSchema  = GameSchema.FromFile(game_id=request.GameID, schema_path=_games_path / request.GameID / "schemas")
    # 1. Get LoaderClass
        load_class = ExportManager._loadLoaderClass(request.GameID)
        if load_class is None:
            # If game doesn't have an extractor, make sure we don't try to export it.
            request.RemoveExportMode(ExportMode.DETECTORS)
            request.RemoveExportMode(ExportMode.SESSION)
            request.RemoveExportMode(ExportMode.PLAYER)
            request.RemoveExportMode(ExportMode.POPULATION)

    # 2. Set up EventManager, assuming it was requested.
        if request.ExportRawEvents or request.ExportProcessedEvents:
            self._event_mgr = EventManager(game_schema=_game_schema, LoaderClass=load_class,
                                           trigger_callback=self._receiveEventTrigger, feature_overrides=request._feat_overrides)
        else:
            Logger.Log("Event data not requested, skipping event manager.", logging.INFO, depth=1)
    # 3. Set up FeatureManager, assuming it was requested.
        if request.ExportSessions or request.ExportPlayers or request.ExportPopulation:
            self._feat_mgr = FeatureManager(game_schema=_game_schema, LoaderClass=load_class, feature_overrides=request._feat_overrides)
        else:
            Logger.Log("Feature data not requested, or extractor loader unavailable, skipping feature manager.", logging.INFO, depth=1)
    # 4. Open the outerfaces
        for outerface in request.Outerfaces:
            outerface.Open()
        self._outputHeaders(request=request)

    def _processSlices(self, request:Request, ids:List[str]) -> None:
        start  : datetime
        slices : List[Slice] = self._generateSlices(sess_ids=ids)

        _next_slice_data : Optional[List[Event]] = None
        for i, next_slice_ids in enumerate(slices):
            _next_slice_data = self._loadSlice(request=request, next_slice_ids=next_slice_ids, slice_num=i+1, slice_count=len(slices))
            if _next_slice_data is not None:
            # 1. Process the slice.
                start = datetime.now()
                Logger.Log(f"Processing slice [{i+1}/{len(slices)}]...", logging.INFO, depth=2)
                self._processSlice(next_slice_data=_next_slice_data, request=request, ids=ids)
                time_delta = datetime.now() - start
                Logger.Log(f"Processing time for slice [{i+1}/{len(slices)}]: {time_delta} to handle {len(_next_slice_data)} events", logging.INFO, depth=2)

            # 2. Write out the session data and reset for next slice.
                start = datetime.now()
                Logger.Log(f"Outputting slice [{i+1}/{len(slices)}]...", logging.INFO, depth=2)
                self._outputSlice(request=request, slice_num=i+1, slice_count=len(slices))
                time_delta = datetime.now() - start
                Logger.Log(f"Output time for slice [{i+1}/{len(slices)}]: {time_delta} to handle {len(_next_slice_data)} events", logging.INFO, depth=2)

    def _postProcess(self, request:Request):
        start = datetime.now()
        if self._feat_mgr is not None:
            if request.ExportPopulation:
                _pop_feats = self._feat_mgr.GetPopulationFeatures(as_str=True)
                for outerface in request.Outerfaces:
                    outerface.WriteLines(lines=_pop_feats, mode=ExportMode.POPULATION)
                self._feat_mgr.ClearPopulationLines()
            if request.ExportPlayers:
                _player_feats = self._feat_mgr.GetPlayerFeatures(as_str=True)
                for outerface in request.Outerfaces:
                    outerface.WriteLines(lines=_player_feats, mode=ExportMode.PLAYER)
                self._feat_mgr.ClearPlayerLines()
        else:
            Logger.Log(f"Skipping feature output for post-process, no FeatureManager exists!", logging.DEBUG, depth=3)
        time_delta = datetime.now() - start
        Logger.Log(f"Output time for population: {time_delta}", logging.INFO, depth=2)

    @staticmethod
    def _loadLoaderClass(game_id:str) -> Optional[Type[GeneratorLoader]]:
        _loader_class: Optional[Type[GeneratorLoader]] = None
        match game_id:
            case "AQUALAB":
                from ogd.games.AQUALAB.AqualabLoader import AqualabLoader
                _loader_class = AqualabLoader
            case "CRYSTAL":
                from ogd.games.CRYSTAL.CrystalLoader import CrystalLoader
                _loader_class = CrystalLoader
            case "ICECUBE":
                from ogd.games.ICECUBE.IcecubeLoader import IcecubeLoader
                _loader_class = IcecubeLoader
            case "JOURNALISM":
                from ogd.games.JOURNALISM.JournalismLoader import JournalismLoader
                _loader_class = JournalismLoader
            case "JOWILDER":
                from ogd.games.JOWILDER.JowilderLoader import JowilderLoader
                _loader_class = JowilderLoader
            case "LAKELAND":
                from ogd.games.LAKELAND.LakelandLoader import LakelandLoader
                _loader_class = LakelandLoader
            case "MAGNET":
                from ogd.games.MAGNET.MagnetLoader import MagnetLoader
                _loader_class = MagnetLoader
            case "SHADOWSPECT":
                from ogd.games.SHADOWSPECT.ShadowspectLoader import ShadowspectLoader
                _loader_class = ShadowspectLoader
            case "SHIPWRECKS":
                from ogd.games.SHIPWRECKS.ShipwrecksLoader import ShipwrecksLoader
                _loader_class = ShipwrecksLoader
            case "THERMOVR":
                from ogd.games.THERMOVR.ThermoVRLoader import ThermoVRLoader
                _loader_class = ThermoVRLoader
            case "WAVES":
                from ogd.games.WAVES.WaveLoader import WaveLoader
                _loader_class = WaveLoader
            case "PENGUINS":
                from ogd.games.PENGUINS.PenguinsLoader import PenguinsLoader
                _loader_class = PenguinsLoader
            case "BLOOM":
                from ogd.games.BLOOM.BloomLoader import BloomLoader
                _loader_class = BloomLoader
            case _:
                if game_id in {"BACTERIA", "BALLOON", "CYCLE_CARBON", "CYCLE_NITROGEN", "CYCLE_WATER", "EARTHQUAKE", "MASHOPOLIS", "WEATHER_STATION", "WIND"}:
                    # all games with data but no extractor.
                    pass
                else:
                    raise ValueError(f"Got an unrecognized game ID ({game_id})!")
        return _loader_class

    def _generateSlices(self, sess_ids:List[str]) -> List[Slice]:
        """
        Convert session ID list into slices of size `self._config.BatchSize.`

        Each slice is just a list of session IDs.
        :param sess_ids: The list of all session IDs in the request.
        :type sess_ids: List[str]
        :return: A list of ID blocks (slices).
        :rtype: List[Slice]
        """
        _num_sess = len(sess_ids)
        _slice_size = self._config.BatchSize
        Logger.Log(f"With slice size = {_slice_size}, there are {math.ceil(_num_sess / _slice_size)} slices", logging.INFO, depth=1)
        return [[sess_ids[i] for i in range( j*_slice_size, min((j+1)*_slice_size, _num_sess) )]
                             for j in range( 0, math.ceil(_num_sess / _slice_size) )]

    def _loadSlice(self, request:Request, next_slice_ids:List[str], slice_num:int, slice_count:int) -> Optional[List[Event]]:
        ret_val : Optional[List[Event]]

        Logger.Log(f"Retrieving slice [{slice_num}/{slice_count}]...", logging.INFO, depth=2)
        start : datetime = datetime.now()
        # TODO : Add a way to configure what to exclude at higher level, here. So we can easily choose to leave out certain events.
        _exclude_rows = None
        # HACK : setting to skip algae events here directly
        if request.GameID == 'BLOOM':
            _exclude_rows = ['algae_growth_end', 'algae_growth_begin']
        ret_val = request.Interface.EventsFromIDs(id_list=next_slice_ids, id_mode=request.Range.IDMode, exclude_rows=_exclude_rows)
        time_delta = datetime.now() - start
        if ret_val is not None:
            Logger.Log(f"Retrieval time for slice [{slice_num}/{slice_count}]: {time_delta} to get {len(ret_val)} events", logging.INFO, depth=2)
        else:
            Logger.Log(f"Could not retrieve data set for slice [{slice_num}/{slice_count}].", logging.WARN, depth=2)
        return ret_val

    def _processSlice(self, next_slice_data:List[Event], request: Request, ids:List[str]):
        _unsessioned_event_count : int = 0
        _sampled_an_event = False
        # 3a) If next slice yielded valid data from the interface, process row-by-row.
        # TODO: instead of separating everything out into one call per event, turn this into a list comprehension using a validation function, so we can pass whole list down a level.
        for event in next_slice_data:
            if not _sampled_an_event:
                Logger.Log(f"First event of slice is:\n{event}", logging.DEBUG, depth=2)
                _sampled_an_event = True
            if (request._range._id_mode==IDMode.SESSION and event.SessionID in ids) \
            or (request._range._id_mode==IDMode.USER    and event.UserID    in ids):
                self._processEvent(next_event=event)
            elif event.SessionID is not None and event.SessionID.upper() != "NONE":
                Logger.Log(f"Found a session ({event.SessionID}, type {type(event.SessionID)}) which was in the slice but not in the list of sessions for processing ({ids[:5]}..., type {type(ids[0])}).", logging.WARNING, depth=2)
            elif event.UserID is not None and event.UserID.upper() != "NONE":
                Logger.Log(f"Found a user ({event.UserID}) which was in the slice but not in the list of players for processing.", logging.WARNING, depth=2)
            else:
                _unsessioned_event_count += 1
                if _unsessioned_event_count < 10:
                    Logger.Log(f"Found an event with no session/player ID, event is: {event}", logging.WARNING, depth=2)
        if _unsessioned_event_count > 0:
            Logger.Log(f"Found {_unsessioned_event_count} events without session IDs.", logging.WARNING, depth=2)

    def _processEvent(self, next_event:Event):
        try:
            if self._event_mgr is not None:
                self._event_mgr.ProcessEvent(event=next_event)
            if self._feat_mgr is not None:
                self._feat_mgr.ProcessEvent(event=next_event)
        except Exception as err:
            if self._config.FailFast:
                Logger.Log(f"Error while processing event {next_event.EventName}:\n{next_event}", logging.ERROR, depth=2)
                raise err
            else:
                Logger.Log(f"Error while processing event {next_event.EventName}. This event will be skipped. \nFull error: {traceback.format_exc()}", logging.WARNING, depth=2)

    def _outputHeaders(self, request:Request):
        if self._event_mgr is not None:
            if request.ExportRawEvents:
                cols = self._event_mgr.GetColumnNames()
                for outerface in request.Outerfaces:
                    outerface.WriteHeader(header=cols, mode=ExportMode.EVENTS)
            else:
                Logger.Log("Event log not requested, skipping events output.", logging.INFO, depth=1)
            if request.ExportProcessedEvents:
                cols = self._event_mgr.GetColumnNames()
                for outerface in request.Outerfaces:
                    outerface.WriteHeader(header=cols, mode=ExportMode.DETECTORS)
            else:
                Logger.Log("Event log not requested, skipping events output.", logging.INFO, depth=1)
        if self._feat_mgr is not None:
            if request.ExportSessions:
                cols = self._feat_mgr.GetSessionFeatureNames()
                for outerface in request.Outerfaces:
                    outerface.WriteHeader(header=cols, mode=ExportMode.SESSION)
            else:
                Logger.Log("Session features not requested, skipping session_features file.", logging.INFO, depth=1)
            if request.ExportPlayers:
                cols = self._feat_mgr.GetPlayerFeatureNames()
                for outerface in request.Outerfaces:
                    outerface.WriteHeader(header=cols, mode=ExportMode.PLAYER)
            else:
                Logger.Log("Player features not requested, skipping player_features file.", logging.INFO, depth=1)
            if request.ExportPopulation:
                cols = self._feat_mgr.GetPopulationFeatureNames()
                for outerface in request.Outerfaces:
                    outerface.WriteHeader(header=cols, mode=ExportMode.POPULATION)
            else:
                Logger.Log("Population features not requested, skipping population_features file.", logging.INFO, depth=1)

    def _outputSlice(self, request:Request, slice_num:int, slice_count:int):
        """
        Output all genearted data for a slice's-worth of raw data

        TODO : Find better solution that does not involve discarding session features, which likely messes up some second-order features.
        :param request: _description_
        :type request: Request
        :param slice_num: _description_
        :type slice_num: int
        :param slice_count: _description_
        :type slice_count: int
        """
        if self._event_mgr is not None:
        # 1. Output raw events, if requested
            if request.ExportRawEvents:
                _events = self._event_mgr.GetRawLines(slice_num=slice_num, slice_count=slice_count)
                for outerface in request.Outerfaces:
                    outerface.WriteLines(lines=_events, mode=ExportMode.EVENTS)
        # 2. Output combined raw & detected events, if requested
            if request.ExportProcessedEvents:
                _events = self._event_mgr.GetAllLines(slice_num=slice_num, slice_count=slice_count)
                for outerface in request.Outerfaces:
                    outerface.WriteLines(lines=_events, mode=ExportMode.DETECTORS)
            self._event_mgr.ClearLines()
        else:
            Logger.Log(f"Skipping event output for slice [{slice_num}/{slice_count}], no EventManager exists!", logging.DEBUG, depth=3)
        if self._feat_mgr is not None:
        # 3. Output session features, if requested
            if request.ExportSessions:
                _sess_feats = self._feat_mgr.GetSessionFeatures(slice_num=slice_num, slice_count=slice_count, as_str=True)
                for outerface in request.Outerfaces:
                    outerface.WriteLines(lines=_sess_feats, mode=ExportMode.SESSION)
                self._feat_mgr.ClearSessionLines()
        else:
            Logger.Log(f"Skipping feature output for slice [{slice_num}/{slice_count}], no FeatureManager exists!", logging.DEBUG, depth=3)



# import libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

# import PerCountyFeature
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.games.BLOOM.features.PerCountyFeature import PerCountyFeature

class CountyLatestMoney(PerCountyFeature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.latest_money: Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["all_events"]
        # return ["switch_county"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        current_money = event.GameState.get("current_money", None)
        self.latest_money = current_money

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.latest_money]

    def Subfeatures(self) -> List[str]:
        return []



# import standard libraries
import abc
import logging
from typing import Any, Dict, List
# import local files
from ogd.core.schemas.games.ExtractorSchema import ExtractorSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class SubfeatureSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, str]):
        self._return_type : str
        self._description : str    

        if not isinstance(all_elements, dict):
            self._elements = {}
            Logger.Log(f"For {name} subfeature config, all_elements was not a dict, defaulting to empty dict", logging.WARN)

        if "return_type" in all_elements.keys():
            self._return_type = SubfeatureSchema._parseReturnType(all_elements['return_type'])
        else:
            self._return_type = "Unknown"
            Logger.Log(f"{name} subfeature config does not have an 'return_type' element; defaulting to return_type='{self._return_type}", logging.WARN)
        if "description" in all_elements.keys():
            self._description = SubfeatureSchema._parseDescription(all_elements['description'])
        else:
            self._description = "No description"
            Logger.Log(f"{name} subfeature config does not have an 'description' element; defaulting to description='{self._description}'", logging.WARN)
        
        _leftovers = { key : val for key,val in all_elements.items() if key not in {"return_type", "description"} }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def ReturnType(self) -> str:
        return self._return_type

    @property
    def Description(self) -> str:
        return self._description

    @property
    def AsMarkdown(self) -> str:
        ret_val : str = f"- **{self.Name}** : *{self.ReturnType}*, {self.Description}  \n"
        if len(self.NonStandardElements) > 0:
            ret_val += f'   (other items: {self.NonStandardElements}'
        return ret_val

    @staticmethod
    def _parseReturnType(return_type):
        ret_val : str
        if isinstance(return_type, str):
            ret_val = return_type
        else:
            ret_val = str(return_type)
            Logger.Log(f"Subfeature return_type was not a string, defaulting to str(return_type) == {ret_val}", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDescription(description):
        ret_val : str
        if isinstance(description, str):
            ret_val = description
        else:
            ret_val = str(description)
            Logger.Log(f"Extractor description was not a string, defaulting to str(description) == {ret_val}", logging.WARN)
        return ret_val

class FeatureSchema(ExtractorSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._subfeatures : Dict[str, SubfeatureSchema]
        self._return_type : str

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Feature config, all_elements was not a dict, defaulting to empty dict", logging.WARN)

        if "return_type" in all_elements.keys():
            self._return_type = FeatureSchema._parseReturnType(all_elements['return_type'], feature_name=name)
        else:
            self._return_type = ""
            Logger.Log(f"{name} Feature config does not have an 'return_type' element; defaulting to return_type='{self._return_type}'", logging.WARN)
        if "subfeatures" in all_elements.keys():
            self._subfeatures = FeatureSchema._parseSubfeatures(all_elements['subfeatures'])
        else:
            self._subfeatures = {}

        _elements = { key : val for key,val in all_elements.items() if key not in {"return_type", "subfeatures"} }
        super().__init__(name=name, all_elements=_elements)

    @property
    def ReturnType(self) -> str:
        return self._return_type

    @property
    def Subfeatures(self) -> Dict[str, SubfeatureSchema]:
        return self._subfeatures

    @staticmethod
    def _parseReturnType(return_type, feature_name:str=""):
        ret_val : str
        if isinstance(return_type, str):
            ret_val = return_type
        else:
            ret_val = str(return_type)
            Logger.Log(f"Feature {feature_name} return_type was not a string, defaulting to str(return_type) == {ret_val}", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSubfeatures(subfeatures) -> Dict[str, SubfeatureSchema]:
        ret_val : Dict[str, SubfeatureSchema]
        if isinstance(subfeatures, dict):
            ret_val = {name:SubfeatureSchema(name=name, all_elements=elems) for name,elems in subfeatures.items()}
        else:
            ret_val = {}
            Logger.Log(f"Extractor subfeatures was unexpected type {type(subfeatures)}, defaulting to empty list.", logging.WARN)
        return ret_val



# import libraries
from datetime import datetime, timedelta
import json
from typing import Any, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

# BUG: quizstart and quizend don't match. Using "wellsdidit" save code to inspect it.

class SurveyTime(PerCountFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params=GeneratorParameters):
        super().__init__(params=params)
        self._start_time : Optional[datetime] = None
        self._duration : Optional[timedelta] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event: Event):
        # NOTE: [0,2,3,4,5] to [0,1,2,3,4]
        quiz_index = event.EventData.get("quiz_number")
        if quiz_index is not None:
            if quiz_index >= 2:
                return quiz_index - 1 == self.CountIndex
            else:
                return quiz_index == self.CountIndex
        else:
            raise KeyError(f"SurveyTime got an event of type {event.EventName} with no quiz_number! EventData keys are: {event.EventData.keys()}")

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.23", "CUSTOM.24"] 
        # ["CUSTOM.23", "CUSTOM.24"] = [quizstart, quizend]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        if self.CountIndex == 0:
            return
        # TODO: Fix the bugs that quizstart and quizend doesn't match
        raise(NotImplementedError("Haven't implemented the function due to bugs"))
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._duration]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return []
    
    @staticmethod
    def MinVersion() -> Optional[str]:
        # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
        return super().MinVersion()

    @staticmethod
    def MaxVersion() -> Optional[str]:
        # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
        return super().MaxVersion()



# import libraries
import json
from typing import Any, Final, List, Optional
import json
from datetime  import timedelta, datetime
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature





class StoryEditorTime(PerLevelFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    IDLE_TIME_THRESHOLD : Final[timedelta] = timedelta(seconds=60)

    def __init__(self, params:GeneratorParameters, threshold: int):
        super().__init__(params=params)
        PerLevelFeature.__init__(self, params=params)


        ##enable feature logic only when in the story editor
        self._editor_watch_enabled: bool = False

        self._first_event_timestamp : Optional[datetime] = None
        self._last_event_timestamp : Optional[datetime] = None

        self._idle_time : Optional[timedelta] = None
        self._time : timedelta = timedelta(0)
        self._count : int = 0
        self._last_timestamp : Optional[datetime] = None
        self._threshold : timedelta = timedelta(seconds=threshold)
        self._total_time : Optional[timedelta] = None

        # >>> create/initialize any variables to track feature extractor state <<<
        #
        # e.g. To track whether extractor found a click event yet:
        # self._found_click : bool = False
    
    @staticmethod
    def defaultThreshold():
        return StoryEditorTime.IDLE_TIME_THRESHOLD

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
        if event.EventName == "open_notebook":
            self._editor_watch_enabled = True
        
        if self._editor_watch_enabled == True:
        
            if not self._first_event_timestamp:
                self._first_event_timestamp = event.Timestamp

            if not self._last_timestamp:
                self._last_timestamp = event.Timestamp
                print("set self._last!")
                return
            if self._last_timestamp is not None:
                time_since_last = event.Timestamp - self._last_timestamp
                if time_since_last > StoryEditorTime.IDLE_TIME_THRESHOLD:
                    self._time += time_since_last
                    self._count += 1
            else:
                raise ValueError("In IdleState, last timestamp is None!")
            self._last_timestamp = event.Timestamp

            if(not self._first_event_timestamp):
                self._first_event_timestamp = event.Timestamp
            
            self._last_event_timestamp = event.Timestamp
        
        ##check to see if editor_watch should be disabled 
        if event.EventName =="close_notebook" and self._editor_watch_enabled == True:
            self._editor_watch_enabled = False


        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        # print(f'_first_timestamp is:{self._first_event_timestamp}, _last_timestamp is: {self._last_timestamp}')
        
        # if(not self._first_event_timestamp):
        #     self._total_time = 0
        #     play_time : int = 0
        # else:
        #     self._total_time = self._last_timestamp-self._first_event_timestamp
        #     play_time : timedelta = self._total_time - self._time

        #total idle time, idle count, total session time, (total session time - idle time)
        # return [self._time, self._count, self._total_time, play_time]
        total_time : timedelta = self._last_event_timestamp - self._first_event_timestamp

        return [total_time, 0, 0, 0]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Count", "Total Time", "Total Play Time"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



# import libraries
import logging
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Type
# import local files
from ogd.core.generators.extractors.builtin.BuiltinExtractor import BuiltinExtractor
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class Timespan(BuiltinExtractor):
    """Universal feature for finding the span of time between the first occurrence of a "start" event, and last occurrence of an "end" event."""
    _start_event = "NO EVENT"
    _end_event   = "NO EVENT"

    def __init__(self, params:GeneratorParameters, schema_args:Dict[str,Any]):
        if params._count_index is not None and params._count_index != 0:
            self.WarningMessage(f"Session feature {params._name} got non-zero count index of {params._count_index}!")
        params._count_index = 0
        super().__init__(params=params, schema_args=schema_args)
        self._start_time : Optional[datetime] = None
        self._end_time   : Optional[datetime] = None
        self._span = timedelta()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _createDerivedGenerator(cls, params:GeneratorParameters, schema_args:Dict[str,Any]) -> Type[BuiltinExtractor]:
        class_params = {
            "_start_event" : schema_args.get("start_event", cls._start_event),
            "_end_event"   : schema_args.get("end_event",   cls._end_event)
        }
        return type(params._name, (Timespan,), class_params)

    @classmethod
    def _eventFilter(cls) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [cls._start_event, cls._end_event]

    @classmethod
    def _featureFilter(cls) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return []

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        match event.EventName:
            case self._start_event:
                if self._start_time is not None:
                    self.WarningMessage(f"{self.Name} received a second {self._start_event} (start-of-span) event! This occurred {event.Timestamp - self._start_time} after the initial {self._start_event} event.")
                else:
                    self._start_time = event.Timestamp
            case self._end_event:
                if self._start_time is not None:
                    if self._end_time is not None:
                        self.WarningMessage(f"{self.Name} received a second {self._end_event} (end-of-span) event! This occurred {event.Timestamp - self._end_time} after the initial {self._end_event} event. Using the later event's timestamp for span.")
                    self._end_time = event.Timestamp
                else:
                    self.WarningMessage(f"{self.Name} received a {self._end_event} event (end-of-span event) when no {self._start_event} event (start-of-span event) had occurred!")
            case _:
                pass
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : List[float]
        if self._start_time is not None and self._end_time is not None:
            ret_val = [(self._end_time - self._start_time).total_seconds()]
        else:
            ret_val = [0]
        return ret_val

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return [] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def MinVersion() -> Optional[str]:
        # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
        return None

    @staticmethod
    def MaxVersion() -> Optional[str]:
        # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
        return None



## import standard libraries
import logging
from collections import OrderedDict
from typing import Any, Callable, List, Optional, Set
## import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import Generator
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.registries.GeneratorRegistry import GeneratorRegistry
from ogd.core.models.Event import Event
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.DetectorSchema import DetectorSchema
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.IterationMode import IterationMode

## @class Extractor
#  Abstract base class for game feature extractors.
#  Gives a few static functions to be used across all extractor classes,
#  and defines an interface that the SessionProcessor can use.
class DetectorRegistry(GeneratorRegistry):
    """Class for registering features to listen for events.

    :return: [description]
    :rtype: [type]
    """

    # *** BUILT-INS & PROPERTIES ***

    # Base constructor for Registry.
    def __init__(self, mode:ExtractionMode, trigger_callback:Callable[[Event], None]):
        """Base constructor for Registry

        Just sets up mostly-empty dictionaries for use by the registry.
        """
        super().__init__(mode=mode)
        self._trigger_callback : Callable[[Event], None]    = trigger_callback
        self._detectors        : OrderedDict[str, Detector] = OrderedDict()

    # string conversion for Extractors.
    def __str__(self) -> str:
        """string conversion for Extractors.

        Creates a list of all features in the extractor, separated by newlines.
        :return: A string with line-separated stringified features.
        :rtype: str
        """
        ret_val : List[str] = []
        ret_val = [str(feat) for feat in self._detectors.values()]
        return '\n'.join(ret_val)

    # Alternate string conversion for Extractors, with limitable number of lines.
    def to_string(self, num_lines:Optional[int] = None) -> str:
        """Alternate string conversion for Extractors, with limitable number of lines.

        Creates a list of features in the extractor, separated by newlines.
        Optional num_lines param allows the function caller to limit number of lines in the string.
        :param num_lines: Max number of lines to include in the string.
        If None, then include all strings, defaults to None
        :type num_lines:  Optional[int], optional
        :return: A string with line-separated stringified features.
        :rtype: str
        """
        ret_val : List[str] = []
        ret_val = [str(feat) for feat in self._detectors.values()]
        if num_lines is None:
            return '\n'.join(ret_val)
        else:
            return '\n'.join(ret_val[:num_lines])

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _register(self, extractor:Generator, iter_mode:IterationMode):
        if isinstance(extractor, Detector):
            _listener = GeneratorRegistry.Listener(name=extractor.Name, mode=iter_mode)
            _event_types   = extractor.EventFilter(mode=self._mode)
            # First, add detector to the _features dict.
            self._detectors[extractor.Name] = extractor
            # Register detector's requested events.
            if "all_events" in _event_types:
                self._event_registry["all_events"].append(_listener)
            else:
                for event in _event_types:
                    if event not in self._event_registry.keys():
                        self._event_registry[event] = []
                    self._event_registry[event].append(_listener)
        else:
            raise TypeError("DetectorRegistry was given an Extractor which was not a Detector!")

    def _getGeneratorNames(self) -> List[str]:
        """Implementation of abstract function to retrieve the names of all extractors currently registered.

        :return: A list of all currently-registered detectors.
        :rtype: List[str]
        """
        ret_val : List[str] = [feature.Name for feature in self._detectors.values()]
        return ret_val

    def _loadFromSchema(self, schema:GameSchema, loader:GeneratorLoader, overrides:Optional[List[str]]=None):
        # first, get list of what should actually be loaded.
        # TODO : move this logic as high up as possible, so that we only need to do it once for each kind of processor.
        # 1. Start with overrides, else list of enabled features in schema.
        agg_load_set : Set[DetectorSchema]
        per_load_set : Set[DetectorSchema]
        if overrides is not None:
            agg_load_set = {schema.Detectors['aggregate'][name] for name in overrides if name in schema.Detectors['aggregate'].keys()}
            per_load_set = {schema.Detectors['per_count'][name] for name in overrides if name in schema.Detectors['per_count'].keys()}
        else:
            agg_load_set = {val for val in schema.EnabledDetectors(iter_modes={IterationMode.AGGREGATE}, extract_modes={self._mode}).values() if isinstance(val, DetectorSchema)}
            per_load_set = {val for val in schema.EnabledDetectors(iter_modes={IterationMode.PERCOUNT}, extract_modes={self._mode}).values() if isinstance(val, DetectorSchema)}
        # 2. Now that we know what needs loading, load them and register.
        # TODO : right now, Detectors are at weird halfway point wrt whether they can be aggregate and percount or not. Need to resolve that.
        detector : Optional[Detector]
        for agg_schema in agg_load_set:
            detector = loader.LoadDetector(detector_type=agg_schema.TypeName, name=agg_schema.Name, schema_args=agg_schema.NonStandardElements, trigger_callback=self._trigger_callback)
            if detector is not None and self._mode in detector.AvailableModes():
                    self.Register(extractor=detector, iter_mode=IterationMode.AGGREGATE)
        for per_schema in per_load_set:
            for i in schema.GetCountRange(count=per_schema.NonStandardElements.get('count', 1)):
                instance_name = f"{per_schema.NonStandardElements.get('prefix', '')}{i}_{per_schema.Name}"
                detector = loader.LoadDetector(detector_type=per_schema.TypeName, name=per_schema.Name, schema_args=per_schema.NonStandardElements, trigger_callback=self._trigger_callback)
                if detector is not None and self._mode in detector.AvailableModes():
                        self.Register(extractor=detector, iter_mode=IterationMode.PERCOUNT)

    def _updateFromEvent(self, event:Event) -> None:
        """Perform extraction of features from a row.

        :param event: [description]
        :type event: Event
        :param table_schema: A data structure containing information on how the db
                             table assiciated with this game is structured.
        :type table_schema: TableSchema
        """
        if event.EventName in self._event_registry.keys():
            # send event to every listener for the given event name.
            for listener in self._event_registry[event.EventName]:
                if listener.name in self._detectors.keys():
                    self._detectors[listener.name].UpdateFromEvent(event)
        # don't forget to send to any features listening for "all" events
        for listener in self._event_registry["all_events"]:
            if listener.name in self._detectors.keys():
                self._detectors[listener.name].UpdateFromEvent(event)

    def _updateFromFeatureData(self, feature:FeatureData) -> None:
        return

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    # def _format(obj):
    #     if obj == None:
    #         return ""
    #     elif type(obj) is timedelta:
    #         total_secs = obj.total_seconds()
    #         # h = total_secs // 3600
    #         # m = (total_secs % 3600) // 60
    #         # s = (total_secs % 3600) % 60 // 1  # just for reference
    #         # return f"{h:02.0f}:{m:02.0f}:{s:02.3f}"
    #         return str(total_secs)
    #     if obj is None:
    #         return ''
    #     else:
    #         return str(obj)



# import libraries
import json
from typing import Any, Dict, List
from datetime import timedelta
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.utils.Logger import Logger
# import libraries
import logging
from ogd.games.PENGUINS.features.bases.PerRegionFeature import PerRegionFeature
    
class RegionDuration(PerRegionFeature):
    
    def __init__(self, params:GeneratorParameters, region_map:List[Dict[str, Any]]):
        super().__init__(params=params,region_map = region_map)
        self._region_start_time = None
        self._time = timedelta(0)
        self._target_region = self._region_map[self.CountIndex].get("name")

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["region_enter","region_exit"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "region_enter":
            self._region_start_time = event.Timestamp
        elif event.EventName == "region_exit":
            if self._region_start_time is not None:
                self._time += (event.Timestamp - self._region_start_time)
                self._region_start_time = None

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._time.total_seconds()]

    def _validateEventCountIndex(self, event:Event):
        _event_region = event.EventData.get("region", "NO REGION FOUND")
        if _event_region == self._target_region:
            return True
        else:
            return False

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Extractor

        Base function to give a list of which ExtractionModes an extractor will handle.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.SESSION]



# import libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class CountyFinalPolicySettings(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.policy_settings: Dict[str, Optional[str]] = {"SalesTaxPolicy": None, "RunoffPolicy": None, "ImportTaxPolicy": None, "SkimmingPolicy": None}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["select_policy_card"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        policy_name = event.EventData.get("policy", None)
        choice_name = event.EventData.get("choice_name", None)

        if policy_name is not None and choice_name is not None:
            if policy_name in self.policy_settings:
                self.policy_settings[policy_name] = choice_name
            else:
                self.WarningMessage(f"Got a select_policy_card with unexpected policy type {policy_name}")

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.policy_settings]

    def Subfeatures(self) -> List[str]:
        return []



## import standard libraries
import abc
import logging
from typing import Any, Dict, List, Optional, Type
# import locals
from ogd.core.generators.Generator import Generator, GeneratorParameters
from ogd.core.generators.extractors.Extractor import Extractor

## @class BuiltinGenerator
class BuiltinExtractor(Extractor):

    # *** ABSTRACTS ***

    ## Abstract function to get a list of event types the Feature wants.
    @classmethod
    @abc.abstractmethod
    def _createDerivedGenerator(cls, params:GeneratorParameters, schema_args:Dict[str,Any]) -> Type["BuiltinExtractor"]:
        """ Abstract function to generate a derived version of the Generator subclass, using the schema args.
            In general, this should just be a matter of creating a type that overwrites the list of events in `_eventFilter` or similar,
            but subclasses are left a lot of leeway in what they change about the derived type.

        :return: [description]
        :rtype: List[str]
        """
        raise TypeError(f"Can't call function on class {cls.__name__} with abstract method _eventFilter")

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters, schema_args:Dict[str,Any]):
        super().__init__(params=params)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import json
from typing import Any, Final, List, Optional
import json
# import local files
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature


"""
QuitType logic:
-"complete_level", "display_feedback", "resumed_checkpoint", "level_fail" all denote certain breakpoints in quit types
-"Buffer_Limit" sets the number of events after a key breakpoint event that the event stays "on"
-every BUFFER_LIMIT events, the stored_event is reset
    -a breakpoint event resets the counter
-time between "display_feedback" event and "complete_level" is considered Quit BetweenLevel
    -counter is set to -1 and does not begin incrementing again until after complete_level
-time between "level_fail" and "resumed_checkpoint"(when origin="LevelFail") is considered Quit OnFail
    -counter is set to -1 and does not begin incrementing again until after resumed_checkpoint

"""


class QuitType(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        
        ##event boolean watchers
        self._store_event: Optional[Event] = None
        self._display_feedback: bool = False
        self._start_level: bool = False
        self._resumed_checkpoint: bool = False
        self._reached_checkpoint: bool = False
        self._level_fail: bool = False

        self._event_watchers : List[str] = ["complete_level", "display_feedback_dialog", "resumed_checkpoint", "level_fail", "reached_checkpoint"]

        ##feature booleans:
        self._between_levels : bool = False
        self._on_fail : bool = False
        self._on_checkpoint : bool = False
        self._other : bool = False

        

        #buffer for event margin before quit type is no longer considered something besides "other"
        self._BUFFER_LIMIT: Final[int] = 9
        #counters            
        self._event_counter : int = 0
    

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if(self._store_event == None):
            self._store_event= event
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
        if(self._event_counter >= self._BUFFER_LIMIT):
            self._event_counter = 0
            self._store_event = event
        elif(self._event_counter != -1 and not('display_' in event.EventName)):
            self._event_counter += 1
        
        if(event.EventName in self._event_watchers):
            #these events have a window of margin longer than BUFFER_LIMIT
            if(event.EventName == "display_feedback_dialog" or event.EventName == "level_fail"):
                self._event_counter=-1
                self._store_event = event

            else:
                self._event_counter = 0
                if(event.EventName=="resumed_checkpoint"):
                    if(event.EventData["origin"]=="LevelFail"):
                        #do NOT overwrite the level_fail in store_event with a resumed_checkpoint from level_fail(preserve level_fail)
                        #start counting with buffer again at this point though
                        pass
                    else:
                        self._store_event = event
                elif(event.EventName=="complete_level"):
                    #do NOT overwrite the display_feedback in store_event
                    #start counting again though
                    pass
                #only store "reached_checkpoint" if we aren't on a new level begin
                elif(not(event.EventName == "reached_checkpoint" and self._store_event != "display_feedback")):
                    self._store_event = event
                    
                else:
                    pass
                    #self._store_event = event
                    


                

        


        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the Feature Data object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        
       

        string_val : Optional[str]
        try:
            if(self._store_event.EventName == "display_feedback_dialog"):
                string_val = "BetweenLevels"
                self._between_levels = True
            elif(self._store_event.EventName == "level_fail"):
                string_val = "OnFail"
                self._on_fail = True
            elif(self._store_event.EventName == "reached_checkpoint"):
                string_val= "OnCheckpoint"
                self._on_checkpoint = True

            else:
                string_val = "Other"
                self._other = True
        except:
            string_val = "null"

        return [string_val, self._between_levels, self._on_fail, self._on_checkpoint, self._other]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["BetweenLevels","OnFail", "OnCheckpoint", "Other"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        ##don't make available for grouping by playerID
        return [ExtractionMode.POPULATION,ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



"""Initializer for Shadowspect features"""

__all__ = [
    "MoveShapeCount",
    "SessionID",
    "FunnelByUser",
    "LevelsOfDifficulty",
    "SequenceBetweenPuzzles",
    "SequenceWithinPuzzles"
]

from . import MoveShapeCount
from . import SessionID
from . import FunnelByUser
from . import LevelsOfDifficulty
from . import SequenceBetweenPuzzles
from . import SequenceWithinPuzzles



from typing import List, Optional, Dict, Any
from models.SequenceModel import SequenceModel
import datetime

## @class TimeSinceLastSaleModel
# Returns the number of seconds since the last purchase a player made
# Rationale: This seems like the simplest and easiest to interpret measure of whether a player
# is stagnating by not selling anything.
# @param levels: Levels applicable for model

class TimeSinceLastSaleModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class TimeSinceLastSaleModel
        Returns the number of seconds since the last purchase a player made
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> Optional[int]:
        assert events
        now = events[-1]["client_time"]
        for event in reversed(events):
            if event["event_custom"] == 37:
                event_time = event["client_time"]
                if type(event_time) is str:
                    event_time = datetime.datetime.fromisoformat(event_time)
                if type(now) is str:
                    now = datetime.datetime.fromisoformat(now)
                return (now - event_time).seconds
        return None


    def __repr__(self):
        return f"TimeSinceLastSaleModel(levels={self._levels}, input_type={self._input_type})"


# import libraries
from typing import Any, List
import json
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class FunnelByUser(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._count = 0
        self._level = None
        self._userFunnelDict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["start_level", "puzzle_started", "create_shape", "check_solution", "puzzle_complete"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName in ["start_level", "puzzle_started"]:
            self._level = event.EventData.get("task_id")
            if isinstance(self._level, dict):
                self._level = self._level["string_value"]

            if self._level not in self._userFunnelDict.keys():
                self._userFunnelDict[self._level] = {"started": 0, "create_shape": 0, "submitted": 0, "completed": 0}

        if self._level != None:
            if event.EventName == "puzzle_started":
                self._userFunnelDict[self._level]["started"] = 1

            elif event.EventName == "create_shape":
                self._userFunnelDict[self._level]["create_shape"] = 1

            elif event.EventName == "check_solution":
                self._userFunnelDict[self._level]["submitted"] = 1

            elif event.EventName == "puzzle_complete":
                self._userFunnelDict[self._level]["completed"] = 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        count = 0
        finalList = []
        listReturn = []
        finalDict = json.loads('{"started": 0, "create_shape": 0, "submitted": 0, "completed": 0}')
        for key in self._userFunnelDict.keys():
            finalList.append((self._userFunnelDict[key]["started"], self._userFunnelDict[key]["create_shape"], self._userFunnelDict[key]["submitted"], self._userFunnelDict[key]["completed"]))
        percent = round((sum([x[3] for x in finalList]) / 30) * 100, 2)
        listReturn.append(percent)
        listReturn.append(sum([x[0] for x in finalList]))
        listReturn.append(sum([x[1] for x in finalList]))
        listReturn.append(sum([x[2] for x in finalList]))
        listReturn.append(sum([x[3] for x in finalList]))
        return listReturn
        
    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["started", "create_shape", "submitted", "completed"]

